"""
Contains AufAbschlagstaffelProOrt class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods


class AufAbschlagstaffelProOrt(COM):
    """
    Gibt den Wert eines Auf- oder Abschlags und dessen Staffelgrenzen an

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlagstaffelProOrt.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlagstaffelProOrt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/AufAbschlagstaffelProOrt.json>`_

    """

    # required attributes
    #: Der Wert für den Auf- oder Abschlag.
    wert: Decimal
    #: Unterer Wert, ab dem die Staffel gilt.
    staffelgrenze_von: Decimal
    #: Oberer Wert, bis zu dem die Staffel gilt.
    staffelgrenze_bis: Decimal



"""
Contains Verbrauch and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import constr, validator

from bo4e.com.com import COM
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.wertermittlungsverfahren import Wertermittlungsverfahren
from bo4e.validators import OBIS_PATTERN, check_bis_is_later_than_von


class Verbrauch(COM):
    """
    Abbildung eines zeitlich abgegrenzten Verbrauchs

    .. raw:: html

        <object data="../_static/images/bo4e/com/Verbrauch.svg" type="image/svg+xml"></object>

    .. HINT::
        `Verbrauch JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Verbrauch.json>`_

    """

    # required attributes
    #: Gibt an, ob es sich um eine PROGNOSE oder eine MESSUNG handelt
    wertermittlungsverfahren: Wertermittlungsverfahren
    #: Die OBIS-Kennzahl für den Wert, die festlegt, welche Größe mit dem Stand gemeldet wird, z.B. '1-0:
    obis_kennzahl: constr(strict=True, regex=OBIS_PATTERN)  # type: ignore[valid-type]  # type: ignore[valid-type]
    #: Gibt den absoluten Wert der Menge an
    wert: Decimal
    #: Gibt die Einheit zum jeweiligen Wert an
    mengeneinheit: Mengeneinheit

    # optional attributes
    #: Inklusiver Beginn des Zeitraumes, für den der Verbrauch angegeben wird
    startdatum: Optional[datetime] = None
    #: Exklusives Ende des Zeitraumes, für den der Verbrauch angegeben wird
    enddatum: Optional[datetime] = None
    _bis_check = validator("enddatum", always=True, allow_reuse=True)(check_bis_is_later_than_von)

    @staticmethod
    def _get_inclusive_start(values: Dict[str, Any]) -> Optional[datetime]:
        """a method for easier usage of the check_bis_is_later_than_von validator"""
        return values["startdatum"]

    # def _get_exclusive_end(self) -> Optional[datetime]:
    #     """a method for easier usage of the check_bis_is_later_than_von validator"""
    #     return self.enddatum



"""
Contains Netznutzungsrechnung class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import constr

from bo4e.bo.rechnung import Rechnung
from bo4e.enum.botyp import BoTyp
from bo4e.enum.nnrechnungsart import NNRechnungsart
from bo4e.enum.nnrechnungstyp import NNRechnungstyp
from bo4e.enum.sparte import Sparte


class Netznutzungsrechnung(Rechnung):
    """
    Modell für die Abbildung von Netznutzungsrechnungen

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Netznutzungsrechnung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Netznutzungsrechnung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Netznutzungsrechnung.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.NETZNUTZUNGSRECHNUNG
    #: Sparte (Strom, Gas ...) für die die Rechnung ausgestellt ist
    sparte: Sparte
    absendercodenummer: constr(strict=True, regex=r"^\d{13}$")  # type: ignore[valid-type]
    """
    Die Rollencodenummer des Absenders (siehe :class:`Marktteilnehmer`).
    Über die Nummer können weitere Informationen zum Marktteilnehmer ermittelt werden.
    """
    empfaengercodenummer: constr(strict=True, regex=r"^\d{13}$")  # type: ignore[valid-type]
    """
    Die Rollencodenummer des Empfängers (siehe :class:`Marktteilnehmer`).
    Über die Nummer können weitere Informationen zum Marktteilnehmer ermittelt werden.
    """
    #: Aus der INVOIC entnommen
    nnrechnungsart: NNRechnungsart
    #: Aus der INVOIC entnommen
    nnrechnungstyp: NNRechnungstyp

    #: Kennzeichen, ob es sich um ein Original (true) oder eine Kopie handelt (false)
    original: bool
    #: Kennzeichen, ob es sich um eine simulierte Rechnung, z.B. zur Rechnungsprüfung handelt
    simuliert: bool

    # optional attributes
    lokations_id: Optional[str] = None
    """
    Die Markt- oder Messlokations-Identifikation (als Malo/Melo-Id) der Lokation, auf die sich die Rechnung bezieht
    """



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.aufabschlagstaffelproort import AufAbschlagstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip


class TestAufAbschlagstaffelProOrt:
    @pytest.mark.parametrize(
        "aufabschlagstaffelproort, expected_json_dict",
        [
            pytest.param(
                AufAbschlagstaffelProOrt(
                    wert=Decimal(2.5),
                    staffelgrenze_von=Decimal(1),
                    staffelgrenze_bis=Decimal(5),
                ),
                {
                    "wert": Decimal("2.5"),
                    "staffelgrenzeVon": Decimal("1"),
                    "staffelgrenzeBis": Decimal("5"),
                },
            ),
        ],
    )
    def test_aufabschlagstaffelproort_required_attributes(
        self, aufabschlagstaffelproort: AufAbschlagstaffelProOrt, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of AufAbschlagstaffelProOrt with minimal attributes.
        """
        assert_serialization_roundtrip(aufabschlagstaffelproort, expected_json_dict)

    def test_aufabschlagstaffelproort_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = AufAbschlagstaffelProOrt()  # type: ignore[call-arg]

        assert "3 validation errors" in str(excinfo.value)



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.kriteriumwert import KriteriumWert
from bo4e.com.regionalegueltigkeit import RegionaleGueltigkeit
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp
from bo4e.enum.tarifregionskriterium import Tarifregionskriterium
from tests.serialization_helper import assert_serialization_roundtrip


class TestRegionaleGueltigkeit:
    @pytest.mark.parametrize(
        "regionalegueltigkeit, expected_json_dict",
        [
            pytest.param(
                RegionaleGueltigkeit(
                    gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
                    kriteriums_werte=[
                        KriteriumWert(
                            kriterium=Tarifregionskriterium.NETZ_NUMMER,
                            wert="12345",
                        ),
                    ],
                ),
                {
                    "gueltigkeitstyp": "NUR_IN",
                    "kriteriumsWerte": [
                        {
                            "kriterium": "NETZ_NUMMER",
                            "wert": "12345",
                        }
                    ],
                },
                id="only required attributes",
            ),
        ],
    )
    def test_regionalegueltigkeit_serialization_roundtrip(
        self, regionalegueltigkeit: RegionaleGueltigkeit, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of RegionaleGueltigkeit with minimal attributes.
        """
        assert_serialization_roundtrip(regionalegueltigkeit, expected_json_dict)

    def test_regionalegueltigkeit_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = RegionaleGueltigkeit()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    def test_regionalegueltigkeit_kriteriumswerte_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = RegionaleGueltigkeit(
                gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
                kriteriums_werte=[],
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


# pylint:disable=empty-docstring
# no docstring in official docs as of 2021-12-01
class Leistungstyp(StrEnum):
    """ """

    ARBEITSPREIS_WIRKARBEIT = "ARBEITSPREIS_WIRKARBEIT"  #: Arbeitspreis zur Abrechnung der Wirkarbeit
    LEISTUNGSPREIS_WIRKLEISTUNG = "LEISTUNGSPREIS_WIRKLEISTUNG"  #: Leistungspreis zur Abrechnung der Wirkleistung
    ARBEITSPREIS_BLINDARBEIT_IND = (
        "ARBEITSPREIS_BLINDARBEIT_IND"  #: Arbeitspreis zur Abrechnung der Blindarbeit induktiv
    )
    ARBEITSPREIS_BLINDARBEIT_KAP = (
        "ARBEITSPREIS_BLINDARBEIT_KAP"  #: Arbeitspreis zur Abrechnung der Blindarbeit kapazitiv
    )
    GRUNDPREIS = "GRUNDPREIS"  #: Grundpreis (pro Zeiteinheit)
    GRUNDPREIS_ARBEIT = "GRUNDPREIS_ARBEIT"  #: Grundpreis, der auf die Arbeit berechnet wird (bei RLM)
    GRUNDPREIS_LEISTUNG = "GRUNDPREIS_LEISTUNG"  #: Grundpreis, der auf die Leistung berechnet wird (bei RLM)
    MEHRMINDERMENGE = "MEHRMINDERMENGE"  #: Mehr- oder Mindermenge
    MESSSTELLENBETRIEB = "MESSSTELLENBETRIEB"  #: Preis pro Zeiteinheit
    MESSDIENSTLEISTUNG = "MESSDIENSTLEISTUNG"  #: Preis pro Zeiteinheit
    MESSDIENSTLEISTUNG_INKL_MESSUNG = (
        "MESSDIENSTLEISTUNG_INKL_MESSUNG"  #: MDL inklusive der Messung (ab 2017), Preis pro Zeiteinheit
    )
    ABRECHNUNG = "ABRECHNUNG"  #: Preis pro Zeiteinheit
    KONZESSIONS_ABGABE = "KONZESSIONS_ABGABE"  #: Konzessionsabgabe
    KWK_UMLAGE = "KWK_UMLAGE"  #: KWK-Umlage
    OFFSHORE_UMLAGE = "OFFSHORE_UMLAGE"  #: Offshore-Haftungsumlage
    ABLAV_UMLAGE = "ABLAV_UMLAGE"  #: Umlage für abschaltbare Lasten
    SONDERKUNDEN_UMLAGE = "SONDERKUNDEN_UMLAGE"  #: §19 StromNEV Umlage
    REGELENERGIE_UMLAGE = "REGELENERGIE_UMLAGE"  #: Regelenergieumlage
    BILANZIERUNG_UMLAGE = "BILANZIERUNG_UMLAGE"  #: Bilanzierungsumlage
    AUSLESUNG_ZUSAETZLICH = "AUSLESUNG_ZUSAETZLICH"  #: Zusätzliche Auslesung (pro Vorgang)
    ABLESUNG_ZUSAETZLICH = "ABLESUNG_ZUSAETZLICH"  #: Zusätzliche Ablesung (pro Vorgang)
    ABRECHNUNG_ZUSAETZLICH = "ABRECHNUNG_ZUSAETZLICH"  #: Zusätzliche Abresung (pro Vorgang)
    SPERRUNG = "SPERRUNG"  #: Sperrung einer Abnahmestelle
    ENTSPERRUNG = "ENTSPERRUNG"  #: Entsperrung einer Abnahmestelle
    MAHNKOSTEN = "MAHNKOSTEN"  #: Mahnkosten
    INKASSOKOSTEN = "INKASSOKOSTEN"  #: Inkassokosten
    EEG_UMLAGE = "EEG_UMLAGE"  #: EEG-Umlage
    ENERGIESTEUER = "ENERGIESTEUER"  #: Strom- oder Erdgassteuer
    NETZPREIS = "NETZPREIS"  #: Netzpreis
    MESSPREIS = "MESSPREIS"  #: Messpreis
    SONSTIGER_PREIS = "SONSTIGER_PREIS"  #: Sonstiger_Preis
    DIENSTLEISTUNG = "DIENSTLEISTUNG"  #: Dienstleistung



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.tarifpreisstaffelproort import TarifpreisstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip

example_tarifpreisstaffelproort = TarifpreisstaffelProOrt(
    arbeitspreis=Decimal(10),
    arbeitspreis_n_t=Decimal(11),
    grundpreis=Decimal(12),
    staffelgrenze_von=Decimal(13),
    staffelgrenze_bis=Decimal(14),
)


class TestTarifpreisstaffelProOrt:
    @pytest.mark.parametrize(
        "tarifpreisstaffelproort",
        [
            pytest.param(
                example_tarifpreisstaffelproort,
                id="maximal (and minimal) attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifpreisstaffelproort: TarifpreisstaffelProOrt) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifpreisstaffelproort)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = TarifpreisstaffelProOrt()  # type: ignore[call-arg]

        assert "5 validation errors" in str(excinfo.value)



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.zeitreihenwertkompakt import Zeitreihenwertkompakt
from bo4e.enum.messwertstatus import Messwertstatus
from bo4e.enum.messwertstatuszusatz import Messwertstatuszusatz


class TestZeitreihenwertkompakt:
    def test_serialization(self) -> None:
        zrwk = Zeitreihenwertkompakt(
            wert=Decimal(1.5), status=Messwertstatus.ABGELESEN, statuszusatz=Messwertstatuszusatz.Z78_GERAETEWECHSEL
        )

        json_string = zrwk.json(by_alias=True, ensure_ascii=False)

        assert "1.5" in json_string
        assert "ABGELESEN" in json_string
        assert "Z78_GERAETEWECHSEL" in json_string
        deserialized_zrwk: Zeitreihenwertkompakt = Zeitreihenwertkompakt.parse_raw(json_string)

        assert isinstance(deserialized_zrwk.wert, Decimal)
        assert deserialized_zrwk.wert == Decimal(1.5)
        assert isinstance(deserialized_zrwk.status, Messwertstatus)
        assert deserialized_zrwk.status == Messwertstatus.ABGELESEN
        assert isinstance(deserialized_zrwk.statuszusatz, Messwertstatuszusatz)
        assert deserialized_zrwk.statuszusatz == Messwertstatuszusatz.Z78_GERAETEWECHSEL
        assert deserialized_zrwk == zrwk

    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitreihenwertkompakt(wert="helloooo")  # type: ignore[arg-type]

        assert "wert" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitreihenwertkompakt(status=Messwertstatus.ABGELESEN)  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)

    def test_only_required(self) -> None:
        zrwk = Zeitreihenwertkompakt(
            wert=Decimal(1.5),
        )

        json_string = zrwk.json(by_alias=True, ensure_ascii=False)

        assert "1.5" in json_string

        deserialized_zrwk: Zeitreihenwertkompakt = Zeitreihenwertkompakt.parse_raw(json_string)

        assert deserialized_zrwk == zrwk



"""
Contains Preisblatt class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.marktteilnehmer import Marktteilnehmer
from bo4e.com.preisposition import Preisposition
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.botyp import BoTyp
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.sparte import Sparte


class Preisblatt(Geschaeftsobjekt):
    """
    Das allgemeine Modell zur Abbildung von Preisen;
    Davon abgeleitet können, über die Zuordnung identifizierender Merkmale, spezielle Preisblatt-Varianten modelliert
    werden.

    Die jeweiligen Sätze von Merkmalen sind in der Grafik ergänzt worden und stellen jeweils eine Ausprägung für die
    verschiedenen Anwendungsfälle der Preisblätter dar.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Preisblatt.svg" type="image/svg+xml"></object>

    .. HINT::
        `Preisblatt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Preisblatt.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.PREISBLATT
    #: Eine Bezeichnung für das Preisblatt
    bezeichnung: str
    #: Preisblatt gilt für angegebene Sparte
    sparte: Sparte
    #: Merkmal, das anzeigt, ob es sich um vorläufige oder endgültige Preise handelt
    preisstatus: Preisstatus
    #: Der Zeitraum für den der Preis festgelegt ist
    gueltigkeit: Zeitraum
    #: Die einzelnen Positionen, die mit dem Preisblatt abgerechnet werden können. Z.B. Arbeitspreis, Grundpreis etc
    preispositionen: conlist(Preisposition, min_items=1)  # type: ignore[valid-type]
    # optional attributes
    #: Der Netzbetreiber, der die Preise veröffentlicht hat
    herausgeber: Optional[Marktteilnehmer] = None



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.betrag import Betrag
from bo4e.com.fremdkostenblock import Fremdkostenblock
from bo4e.com.fremdkostenposition import Fremdkostenposition
from bo4e.com.preis import Preis
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.waehrungscode import Waehrungscode
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip


class TestFremdkostenblock:
    @pytest.mark.parametrize(
        "fremdkostenblock, expected_json_dict",
        [
            pytest.param(
                Fremdkostenblock(
                    kostenblockbezeichnung="teststring",
                    kostenpositionen=[
                        Fremdkostenposition(
                            positionstitel="fremdkostenblocktitel",
                            artikelbezeichnung="bsp",
                            einzelpreis=Preis(
                                wert=Decimal(3.50),
                                einheit=Waehrungseinheit.EUR,
                                bezugswert=Mengeneinheit.KWH,
                                status=Preisstatus.ENDGUELTIG,
                            ),
                            betrag_kostenposition=Betrag(
                                waehrung=Waehrungscode.EUR,
                                wert=Decimal(12.5),
                            ),
                        )
                    ],
                    summe_kostenblock=Betrag(
                        waehrung=Waehrungscode.EUR,
                        wert=Decimal(98240),
                    ),
                ),
                {
                    "kostenblockbezeichnung": "teststring",
                    "kostenpositionen": [
                        {
                            "marktpartnercode": None,
                            "positionstitel": "fremdkostenblocktitel",
                            "einzelpreis": {
                                "wert": Decimal("3.5"),
                                "einheit": Waehrungseinheit.EUR,
                                "bezugswert": Mengeneinheit.KWH,
                                "status": Preisstatus.ENDGUELTIG,
                            },
                            "bis": None,
                            "menge": None,
                            "zeitmenge": None,
                            "artikelbezeichnung": "bsp",
                            "marktpartnername": None,
                            "artikeldetail": None,
                            "von": None,
                            "linkPreisblatt": None,
                            "betragKostenposition": {"wert": Decimal("12.5"), "waehrung": Waehrungseinheit.EUR},
                            "gebietcodeEic": None,
                        }
                    ],
                    "summeKostenblock": {"wert": Decimal("98240"), "waehrung": Waehrungseinheit.EUR},
                },
                id="maximal attributes",
            ),
            pytest.param(
                Fremdkostenblock(kostenblockbezeichnung="teststring"),
                {
                    "kostenblockbezeichnung": "teststring",
                    "kostenpositionen": None,
                    "summeKostenblock": None,
                },
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, fremdkostenblock: Fremdkostenblock, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Fremdkostenblock
        """
        assert_serialization_roundtrip(fremdkostenblock, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Fremdkostenblock()  # type: ignore[call-arg]
        assert "1 validation error" in str(excinfo.value)



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.sigmoidparameter import Sigmoidparameter
from tests.serialization_helper import assert_serialization_roundtrip

# this sigmoid parameter can be imported by other tests
example_sigmoidparameter = Sigmoidparameter(
    A=Decimal(1),
    B=Decimal(2),
    C=Decimal(3),
    D=Decimal(4),
)


class TestSigmoidparameter:
    @pytest.mark.parametrize(
        "sigmoidparameter, expected_json_dict",
        [
            pytest.param(
                example_sigmoidparameter,
                {"A": Decimal("1"), "B": Decimal("2"), "C": Decimal("3"), "D": Decimal("4")},
            ),
        ],
    )
    def test_sigmoidparameter_serialization_roundtrip(
        self, sigmoidparameter: Sigmoidparameter, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Sigmoidparameter with minimal attributes.
        """
        assert_serialization_roundtrip(sigmoidparameter, expected_json_dict)

    def test_sigmoidparameter_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Sigmoidparameter()  # type: ignore[call-arg]
        assert "4 validation errors" in str(excinfo.value)

    @pytest.mark.parametrize(
        "sigmoidparameter, leistung, expected_lp",
        [
            pytest.param(
                Sigmoidparameter(
                    A=Decimal(1),
                    B=Decimal(3),
                    C=Decimal(3),
                    D=Decimal(4),
                ),
                Decimal(3),
                Decimal(4.5),
            ),
        ],
    )
    def test_lp_calculation(self, sigmoidparameter: Sigmoidparameter, leistung: Decimal, expected_lp: Decimal) -> None:
        assert sigmoidparameter.calculate(leistung) == expected_lp



from datetime import datetime
from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.bo.zaehler import Zaehler
from bo4e.com.externereferenz import ExterneReferenz
from bo4e.com.zaehlwerk import Zaehlwerk
from bo4e.enum.botyp import BoTyp
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.zaehlerauspraegung import Zaehlerauspraegung
from bo4e.enum.zaehlertyp import Zaehlertyp


class TestZaehler:
    def test_de_serialisation(self) -> None:
        """
        Test de-/serialisation of Zaehler only with required attributes
        """
        zaehler = Zaehler(
            zaehlernummer="000111222",
            sparte=Sparte.STROM,
            zaehlerauspraegung=Zaehlerauspraegung.EINRICHTUNGSZAEHLER,
            zaehlwerke=[
                Zaehlwerk(
                    zaehlwerk_id="98765",
                    einheit=Mengeneinheit.KW,
                    richtung=Energierichtung.EINSP,
                    bezeichnung="my zaehlwerk",
                    obis_kennzahl="1-0:1.8.1",
                    wandlerfaktor=Decimal(0.95),
                )
            ],
            zaehlertyp=Zaehlertyp.DREHSTROMZAEHLER,
            tarifart=Tarifart.ZWEITARIF,
            zaehlerkonstante=Decimal(0.9),
            eichung_bis=datetime(2022, 1, 1, 0, 0, 0),
            externe_referenzen=[ExterneReferenz(ex_ref_name="zaehler im anderen system", ex_ref_wert="7890")],
            letzte_eichung=datetime(2019, 6, 30, 0, 0, 0),
        )
        assert zaehler.versionstruktur == "2", "versionstruktur was not automatically set"
        assert zaehler.bo_typ is BoTyp.ZAEHLER, "boTyp was not automatically set"
        assert zaehler.zaehlwerke[0].richtung == Energierichtung.EINSP
        assert zaehler.zaehlwerke[0].einheit == Mengeneinheit.KW
        json_string = zaehler.json(by_alias=True, ensure_ascii=False)
        assert "richtung" in json_string, "Zaehlwerk->richtung was not serialized"
        assert "einheit" in json_string, "Zaehlwerk->einheit was not serialized"
        deserialized_zaehler = Zaehler.parse_raw(json_string)
        assert deserialized_zaehler == zaehler

    def test_serialization_fails_for_invalid_obis(self) -> None:
        """
        Test serialisation of Zaehler fails if OBIS is wrong.
        """
        with pytest.raises(ValidationError) as excinfo:
            _ = Zaehler(
                zaehlernummer="000111222",
                sparte=Sparte.STROM,
                zaehlerauspraegung=Zaehlerauspraegung.EINRICHTUNGSZAEHLER,
                zaehlwerke=[
                    Zaehlwerk(
                        zaehlwerk_id="98765",
                        einheit=Mengeneinheit.KW,
                        richtung=Energierichtung.EINSP,
                        bezeichnung="my zaehlwerk",
                        obis_kennzahl="foo",  # <-- this is not a valid obis. it triggered the value error
                        wandlerfaktor=Decimal(0.95),
                    )
                ],
                zaehlertyp=Zaehlertyp.DREHSTROMZAEHLER,
                tarifart=Tarifart.ZWEITARIF,
            )
        assert "1 validation error" in str(excinfo.value)
        assert "obisKennzahl" in str(excinfo.value)
        assert "string does not match regex" in str(excinfo.value)

    def test_serialization_fails_for_empty_zaehlwerke(self) -> None:
        """
        Test serialisation of Zaehler fails if there are no zaehlwerke.
        """
        with pytest.raises(ValidationError) as excinfo:
            _ = Zaehler(
                zaehlernummer="000111222",
                sparte=Sparte.STROM,
                zaehlerauspraegung=Zaehlerauspraegung.EINRICHTUNGSZAEHLER,
                zaehlwerke=[],
                zaehlertyp=Zaehlertyp.DREHSTROMZAEHLER,
                tarifart=Tarifart.ZWEITARIF,
            )
        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.aufabschlag import AufAbschlag
from bo4e.com.preisstaffel import Preisstaffel
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.aufabschlagsziel import AufAbschlagsziel
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_sigmoidparameter import example_sigmoidparameter

example_aufabschlag = AufAbschlag(
    bezeichnung="foo",
    staffeln=[
        Preisstaffel(
            einheitspreis=Decimal(15.0),
            staffelgrenze_von=Decimal(2.5),
            staffelgrenze_bis=Decimal(40.5),
        ),
    ],
)


class TestAufAbschlag:
    @pytest.mark.parametrize(
        "aufabschlag, expected_json_dict",
        [
            pytest.param(
                AufAbschlag(
                    bezeichnung="foo",
                    beschreibung="bar",
                    auf_abschlagstyp=AufAbschlagstyp.ABSOLUT,
                    auf_abschlagsziel=AufAbschlagsziel.GESAMTPREIS,
                    einheit=Waehrungseinheit.EUR,
                    website="foo.bar",
                    gueltigkeitszeitraum=Zeitraum(
                        startdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
                        enddatum=datetime(2020, 4, 1, tzinfo=timezone.utc),
                    ),
                    staffeln=[
                        Preisstaffel(
                            einheitspreis=Decimal(40.0),
                            staffelgrenze_von=Decimal(12.5),
                            staffelgrenze_bis=Decimal(25.0),
                            sigmoidparameter=example_sigmoidparameter,
                        ),
                        Preisstaffel(
                            einheitspreis=Decimal(15.0),
                            staffelgrenze_von=Decimal(2.5),
                            staffelgrenze_bis=Decimal(40.5),
                        ),
                    ],
                ),
                {
                    "bezeichnung": "foo",
                    "beschreibung": "bar",
                    "aufAbschlagstyp": AufAbschlagstyp.ABSOLUT,
                    "aufAbschlagsziel": AufAbschlagsziel.GESAMTPREIS,
                    "einheit": Waehrungseinheit.EUR,
                    "website": "foo.bar",
                    "gueltigkeitszeitraum": {
                        "startdatum": datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
                        "endzeitpunkt": None,
                        "einheit": None,
                        "enddatum": datetime(2020, 4, 1, 0, 0, tzinfo=timezone.utc),
                        "startzeitpunkt": None,
                        "dauer": None,
                    },
                    "staffeln": [
                        {
                            "einheitspreis": Decimal("40"),
                            "sigmoidparameter": {
                                "A": Decimal("1"),
                                "B": Decimal("2"),
                                "C": Decimal("3"),
                                "D": Decimal("4"),
                            },
                            "staffelgrenzeVon": Decimal("12.5"),
                            "staffelgrenzeBis": Decimal("25"),
                        },
                        {
                            "einheitspreis": Decimal("15"),
                            "sigmoidparameter": None,
                            "staffelgrenzeVon": Decimal("2.5"),
                            "staffelgrenzeBis": Decimal("40.5"),
                        },
                    ],
                },
                id="maximal attributes",
            ),
            pytest.param(
                example_aufabschlag,
                {
                    "bezeichnung": "foo",
                    "beschreibung": None,
                    "aufAbschlagstyp": None,
                    "aufAbschlagsziel": None,
                    "einheit": None,
                    "website": None,
                    "gueltigkeitszeitraum": None,
                    "staffeln": [
                        {
                            "einheitspreis": Decimal("15"),
                            "sigmoidparameter": None,
                            "staffelgrenzeVon": Decimal("2.5"),
                            "staffelgrenzeBis": Decimal("40.5"),
                        },
                    ],
                },
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, aufabschlag: AufAbschlag, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of AufAbschlag with minimal attributes.
        """
        assert_serialization_roundtrip(aufabschlag, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = AufAbschlag()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    @pytest.mark.parametrize(
        "auf_abschlagstyp",
        [
            pytest.param(
                AufAbschlagstyp.RELATIV,
                id="auf_abschlagstyp not absolute",
            ),
            pytest.param(
                None,
                id="auf_abschlagstyp not there",
            ),
        ],
    )
    def test_failing_validation_einheit_only_for_abschlagstyp_absolut(self, auf_abschlagstyp: AufAbschlagstyp) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = AufAbschlag(
                bezeichnung="foo",
                staffeln=[
                    Preisstaffel(
                        einheitspreis=Decimal(15.0),
                        staffelgrenze_von=Decimal(2.5),
                        staffelgrenze_bis=Decimal(40.5),
                    ),
                ],
                einheit=Waehrungseinheit.EUR,
                auf_abschlagstyp=auf_abschlagstyp,
            )

        assert "Only state einheit if auf_abschlagstyp is absolute." in str(excinfo.value)



from datetime import datetime, timezone
from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.bo.netznutzungsrechnung import Netznutzungsrechnung
from bo4e.com.betrag import Betrag
from bo4e.com.rechnungsposition import Rechnungsposition
from bo4e.com.steuerbetrag import Steuerbetrag
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.nnrechnungsart import NNRechnungsart
from bo4e.enum.nnrechnungstyp import NNRechnungstyp
from bo4e.enum.rechnungsstatus import Rechnungsstatus
from bo4e.enum.rechnungstyp import Rechnungstyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.steuerkennzeichen import Steuerkennzeichen
from bo4e.enum.waehrungscode import Waehrungscode
from bo4e.enum.zeiteinheit import Zeiteinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_betrag import example_betrag
from tests.test_menge import example_menge
from tests.test_preis import example_preis
from tests.test_steuerbetrag import example_steuerbetrag

_rechnungsersteller = Geschaeftspartner(
    name1="Joachim",
    gewerbekennzeichnung=False,
    geschaeftspartnerrolle=[Geschaeftspartnerrolle.LIEFERANT],
)

_rechnungsempfaenger = Geschaeftspartner(
    name1="Helga",
    gewerbekennzeichnung=False,
    geschaeftspartnerrolle=[Geschaeftspartnerrolle.KUNDE],
)


class TestNetznutzungsrechnung:
    @pytest.mark.parametrize(
        "nnrechnung",
        [
            pytest.param(
                Netznutzungsrechnung(
                    sparte=Sparte.STROM,
                    absendercodenummer="9876543210123",
                    empfaengercodenummer="0123456789012",
                    nnrechnungsart=NNRechnungsart.SELBSTAUSGESTELLT,
                    nnrechnungstyp=NNRechnungstyp.TURNUSRECHNUNG,
                    original=True,
                    simuliert=True,
                    lokations_id="56789012345",
                    # ^^ above are the original Netznutzungsrechnung attributes
                    # vv below are the fields inherited from Rechnung
                    rechnungstitel="Hüpfburg",
                    rechnungsstatus=Rechnungsstatus.UNGEPRUEFT,
                    storno=True,
                    rechnungsnummer="202201211701",
                    rechnungsdatum=datetime.today(),
                    faelligkeitsdatum=datetime.today(),
                    rechnungstyp=Rechnungstyp.ENDKUNDENRECHNUNG,
                    original_rechnungsnummer="RE-2022-01-21_1701",
                    rechnungsperiode=Zeitraum(einheit=Zeiteinheit.TAG, dauer=Decimal(21)),
                    rechnungsersteller=_rechnungsersteller,
                    rechnungsempfaenger=_rechnungsempfaenger,
                    gesamtnetto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtsteuer=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtbrutto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    vorausgezahlt=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    rabatt_brutto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    zuzahlen=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    steuerbetraege=[
                        Steuerbetrag(
                            steuerkennzeichen=Steuerkennzeichen.UST_19,
                            basiswert=Decimal(20.25),
                            waehrung=Waehrungscode.EUR,
                            steuerwert=Decimal(10.5),
                        )
                    ],
                    rechnungspositionen=[
                        Rechnungsposition(
                            positionsnummer=1,
                            lieferung_von=datetime(2021, 3, 15, tzinfo=timezone.utc),
                            lieferung_bis=datetime(2022, 3, 15, tzinfo=timezone.utc),
                            positionstext="Besonders wertvolle Rechnungsposition",
                            positions_menge=example_menge,
                            einzelpreis=example_preis,
                            teilsumme_netto=example_betrag,
                            teilsumme_steuer=example_steuerbetrag,
                        )
                    ],
                ),
                id="maximal attributes",
            ),
            pytest.param(
                Netznutzungsrechnung(
                    sparte=Sparte.STROM,
                    absendercodenummer="9876543210123",
                    empfaengercodenummer="0123456789012",
                    nnrechnungsart=NNRechnungsart.SELBSTAUSGESTELLT,
                    nnrechnungstyp=NNRechnungstyp.TURNUSRECHNUNG,
                    original=True,
                    simuliert=True,
                    # ^^ above are the original Netznutzungsrechnung attributes
                    # vv below are the fields inherited from Rechnung
                    storno=True,
                    rechnungsnummer="202201211701",
                    rechnungsdatum=datetime.today(),
                    faelligkeitsdatum=datetime.today(),
                    rechnungstyp=Rechnungstyp.ENDKUNDENRECHNUNG,
                    original_rechnungsnummer="RE-2022-01-21_1701",
                    rechnungsperiode=Zeitraum(einheit=Zeiteinheit.TAG, dauer=Decimal(21)),
                    rechnungsersteller=_rechnungsersteller,
                    rechnungsempfaenger=_rechnungsempfaenger,
                    gesamtnetto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtsteuer=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    gesamtbrutto=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    zuzahlen=Betrag(wert=Decimal(12.5), waehrung=Waehrungscode.EUR),
                    rechnungspositionen=[
                        Rechnungsposition(
                            positionsnummer=1,
                            lieferung_von=datetime(2021, 3, 15, tzinfo=timezone.utc),
                            lieferung_bis=datetime(2022, 3, 15, tzinfo=timezone.utc),
                            positionstext="Besonders wertvolle Rechnungsposition",
                            positions_menge=example_menge,
                            einzelpreis=example_preis,
                            teilsumme_netto=example_betrag,
                            teilsumme_steuer=example_steuerbetrag,
                        )
                    ],
                ),
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, nnrechnung: Netznutzungsrechnung) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(nnrechnung)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Netznutzungsrechnung()  # type: ignore[call-arg]
        assert "20 validation errors" in str(excinfo.value)  # 13 from rechnung + 7 from Netznutzungsrechnung



"""
Contains Rechnung class
and corresponding marshmallow schema for de-/serialization
"""

from datetime import datetime
from typing import List, Optional

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.betrag import Betrag
from bo4e.com.rechnungsposition import Rechnungsposition
from bo4e.com.steuerbetrag import Steuerbetrag
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.botyp import BoTyp
from bo4e.enum.rechnungsstatus import Rechnungsstatus
from bo4e.enum.rechnungstyp import Rechnungstyp

# pylint: disable=too-few-public-methods, too-many-instance-attributes


class Rechnung(Geschaeftsobjekt):
    """
    Modell für die Abbildung von Rechnungen im Kontext der Energiewirtschaft;
    Ausgehend von diesem Basismodell werden weitere spezifische Formen abgeleitet.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Rechnung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Rechnung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Rechnung.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.RECHNUNG
    storno: bool
    """
    Kennzeichnung, ob es sich um eine Stornorechnung handelt;
    im Falle "true" findet sich im Attribut "originalrechnungsnummer" die Nummer der Originalrechnung.
    """
    #: Eine im Verwendungskontext eindeutige Nummer für die Rechnung
    rechnungsnummer: str
    #: Ausstellungsdatum der Rechnung
    rechnungsdatum: datetime
    #: Zu diesem Datum ist die Zahlung fällig
    faelligkeitsdatum: datetime
    #: Ein kontextbezogender Rechnungstyp, z.B. Netznutzungsrechnung
    rechnungstyp: Rechnungstyp
    #: Der Zeitraum der zugrunde liegenden Lieferung zur Rechnung
    rechnungsperiode: Zeitraum
    #: Der Aussteller der Rechnung
    rechnungsersteller: Geschaeftspartner
    #: Der Aussteller der Rechnung
    rechnungsempfaenger: Geschaeftspartner
    #: Die Summe der Nettobeträge der Rechnungsteile
    gesamtnetto: Betrag
    #: Die Summe der Steuerbeträge der Rechnungsteile
    gesamtsteuer: Betrag
    #: Die Summe aus Netto- und Steuerbetrag
    gesamtbrutto: Betrag
    #: Der zu zahlende Betrag, der sich aus (gesamtbrutto - vorausbezahlt - rabattBrutto) ergibt
    zuzahlen: Betrag
    #: Die Rechnungspositionen
    rechnungspositionen: List[Rechnungsposition]

    # optional attributes
    #: Bezeichnung für die vorliegende Rechnung
    rechnungstitel: Optional[str] = None
    #: Status der Rechnung zur Kennzeichnung des Bearbeitungsstandes
    rechnungsstatus: Optional[Rechnungsstatus] = None
    #: Im Falle einer Stornorechnung (storno = true) steht hier die Rechnungsnummer der stornierten Rechnung
    original_rechnungsnummer: Optional[str] = None
    #: Die Summe evtl. vorausgezahlter Beträge, z.B. Abschläge. Angabe als Bruttowert
    vorausgezahlt: Optional[Betrag] = None
    #: Gesamtrabatt auf den Bruttobetrag
    rabatt_brutto: Optional[Betrag] = None
    steuerbetraege: Optional[List[Steuerbetrag]] = None
    """
    Eine Liste mit Steuerbeträgen pro Steuerkennzeichen/Steuersatz;
    die Summe dieser Beträge ergibt den Wert für gesamtsteuer.
    """



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Verbrauchsart(StrEnum):
    """
    Verbrauchsart einer Marktlokation.
    """

    KL = "KL"  #: Kraft/Licht
    KLW = "KLW"  #: Kraft/Licht/Wärme
    KLWS = "KLWS"  #: Kraft/Licht/Wärme/Speicherheizung
    W = "W"  #: Wärme
    WS = "WS"  #: Wärme/Speicherheizung



import pytest
from pydantic import ValidationError

from bo4e.bo.preisblattmessung import PreisblattMessung
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.sparte import Sparte
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_geraeteeigenschaften import example_geraeteeigenschaften
from tests.test_marktteilnehmer import example_marktteilnehmer
from tests.test_preisposition import example_preisposition
from tests.test_zeitraum import example_zeitraum


class TestPreisblattMessung:
    @pytest.mark.parametrize(
        "preisblatt_messung",
        [
            pytest.param(
                PreisblattMessung(
                    bezeichnung="foo",
                    sparte=Sparte.STROM,
                    preisstatus=Preisstatus.ENDGUELTIG,
                    preispositionen=[example_preisposition],
                    gueltigkeit=example_zeitraum,
                    herausgeber=example_marktteilnehmer,
                    bilanzierungsmethode=Bilanzierungsmethode.TLP_GEMEINSAM,
                    messebene=Netzebene.MSP,
                    inklusive_dienstleistungen=[Dienstleistungstyp.AUSLESUNG_FERNAUSLESUNG_ZUSAETZLICH_MSB],
                    zaehler=example_geraeteeigenschaften,
                    inklusive_geraete=[example_geraeteeigenschaften],
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, preisblatt_messung: PreisblattMessung) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(preisblatt_messung)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = PreisblattMessung()  # type: ignore[call-arg]
        assert "8 validation errors" in str(excinfo.value)  # 5 from preisblatt + 3 from preisblatt messung



"""
Contains ExterenzeReferenz class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM


# pylint: disable=too-few-public-methods
#
class ExterneReferenz(COM):
    """
    Viele Datenobjekte weisen in unterschiedlichen Systemen eine eindeutige ID (Kundennummer, GP-Nummer etc.) auf.
    Beim Austausch von Datenobjekten zwischen verschiedenen Systemen ist es daher hilfreich,
    sich die eindeutigen IDs der anzubindenden Systeme zu merken.

    .. raw:: html

        <object data="../_static/images/bo4e/com/ExterneReferenz.svg" type="image/svg+xml"></object>

    .. HINT::
        `ExterneReferenz JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/ExterneReferenz.json>`_

    """

    # required attributes
    ex_ref_name: str  #: Bezeichnung der externen Referenz (z.B. "microservice xyz" oder "SAP CRM GP-Nummer")
    ex_ref_wert: str  #: Wert der externen Referenz (z.B. "123456"; "4711")



"""
Contains Marktlokation class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-many-instance-attributes, too-few-public-methods
from typing import Any, Dict, Optional

# pylint: disable=no-name-in-module
from pydantic import conlist, validator

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse
from bo4e.com.geokoordinaten import Geokoordinaten
from bo4e.com.katasteradresse import Katasteradresse
from bo4e.com.messlokationszuordnung import Messlokationszuordnung
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.gasqualitaet import Gasqualitaet
from bo4e.enum.gebiettyp import Gebiettyp
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte
from bo4e.enum.verbrauchsart import Verbrauchsart
from bo4e.validators import validate_marktlokations_id


class Marktlokation(Geschaeftsobjekt):
    """
    Object containing information about a Marktlokation

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Marktlokation.svg" type="image/svg+xml"></object>

    .. HINT::
        `Marktlokation JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Marktlokation.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.MARKTLOKATION
    #: Identifikationsnummer einer Marktlokation, an der Energie entweder verbraucht, oder erzeugt wird.
    marktlokations_id: str
    _marktlokations_id_check = validator("marktlokations_id", allow_reuse=True)(validate_marktlokations_id)
    #: Sparte der Marktlokation, z.B. Gas oder Strom
    sparte: Sparte
    #: Kennzeichnung, ob Energie eingespeist oder entnommen (ausgespeist) wird
    energierichtung: Energierichtung
    #: Die Bilanzierungsmethode, RLM oder SLP
    bilanzierungsmethode: Bilanzierungsmethode
    netzebene: Netzebene
    """
    Netzebene, in der der Bezug der Energie erfolgt.
    Bei Strom Spannungsebene der Lieferung, bei Gas Druckstufe.
    Beispiel Strom: Niederspannung Beispiel Gas: Niederdruck.
    """

    # optional attributes
    #: Verbrauchsart der Marktlokation.
    verbrauchsart: Optional[Verbrauchsart] = None
    #: Gibt an, ob es sich um eine unterbrechbare Belieferung handelt
    unterbrechbar: Optional[bool] = None
    #: Codenummer des Netzbetreibers, an dessen Netz diese Marktlokation angeschlossen ist.
    netzbetreibercodenr: Optional[str] = None
    #: Typ des Netzgebietes, z.B. Verteilnetz
    gebietstyp: Optional[Gebiettyp] = None
    #: Die ID des Gebietes in der ene't-Datenbank
    netzgebietsnr: Optional[str] = None  # todo: rename to "id" (see 2021-12-15 update)
    #: Bilanzierungsgebiet, dem das Netzgebiet zugeordnet ist - im Falle eines Strom Netzes
    bilanzierungsgebiet: Optional[str] = None
    #: Codenummer des Grundversorgers, der für diese Marktlokation zuständig ist
    grundversorgercodenr: Optional[str] = None
    #: Die Gasqualität in diesem Netzgebiet. H-Gas oder L-Gas. Im Falle eines Gas-Netzes
    gasqualitaet: Optional[Gasqualitaet] = None
    #: Geschäftspartner, dem diese Marktlokation gehört
    endkunde: Optional[Geschaeftspartner] = None
    zugehoerige_messlokation: Optional[Messlokationszuordnung] = None  # todo: rename to plural
    """
    Aufzählung der Messlokationen, die zu dieser Marktlokation gehören.
    Es können 3 verschiedene Konstrukte auftreten:

    Beziehung 1 : 0 : Hier handelt es sich um Pauschalanlagen ohne Messung. D.h. die Verbrauchsdaten sind direkt über
    die Marktlokation abgreifbar.
    Beziehung 1 : 1 : Das ist die Standard-Beziehung für die meisten Fälle. In diesem Fall gibt es zu einer
    Marktlokation genau eine Messlokation.
    Beziehung 1 : N : Hier liegt beispielsweise eine Untermessung vor. Der Verbrauch einer Marklokation berechnet sich
    hier aus mehreren Messungen.

    Es gibt praktisch auch noch die Beziehung N : 1, beispielsweise bei einer Zweirichtungsmessung bei der durch eine
    Messeinrichtung die Messung sowohl für die Einspreiseseite als auch für die Aussspeiseseite erfolgt.
    Da Abrechnung und Bilanzierung jedoch für beide Marktlokationen getrennt erfolgt, werden nie beide Marktlokationen
    gemeinsam betrachtet. Daher lässt sich dieses Konstrukt auf zwei 1:1-Beziehung zurückführen,
    wobei die Messlokation in beiden Fällen die gleiche ist.

    In den Zuordnungen sind ist die arithmetische Operation mit der der Verbrauch einer Messlokation zum Verbrauch einer
    Marktlokation beitrögt mit aufgeführt.
    Der Standard ist hier die Addition.
    """

    # only one of the following three optional attributes can be set
    #: Die Adresse, an der die Energie-Lieferung oder -Einspeisung erfolgt
    lokationsadresse: Optional[Adresse] = None
    geoadresse: Optional[Geokoordinaten] = None
    """
    Alternativ zu einer postalischen Adresse kann hier ein Ort mittels Geokoordinaten angegeben werden
    (z.B. zur Identifikation von Sendemasten).
    """
    katasterinformation: Optional[Katasteradresse] = None
    """
    Alternativ zu einer postalischen Adresse und Geokoordinaten kann hier eine Ortsangabe mittels Gemarkung und
    Flurstück erfolgen.
    """

    kundengruppen: conlist(Kundentyp, min_items=0) = None  # type: ignore[valid-type]
    #: Kundengruppen der Marktlokation

    # pylint:disable=unused-argument, no-self-argument
    @validator("katasterinformation", always=True)
    def validate_address_info(
        cls, katasterinformation: Optional[Katasteradresse], values: Dict[str, Any]
    ) -> Optional[Katasteradresse]:
        """Checks that there is one and only one valid adress given."""
        all_address_attributes = [
            values["lokationsadresse"],
            values["geoadresse"],
            katasterinformation,
        ]
        amount_of_given_address_infos = len([i for i in all_address_attributes if i is not None])
        if amount_of_given_address_infos != 1:
            raise ValueError("No or more than one address information is given.")
        return katasterinformation



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Ausschreibungstyp(StrEnum):
    """
    Aufzählung für die Typisierung von Ausschreibungen.
    """

    PRIVATRECHTLICH = "PRIVATRECHTLICH"  #: privat-rechtlich
    OEFFENTLICHRECHTLICH = "OEFFENTLICHRECHTLICH"  #: öffentlich-rechtlich
    EUROPAWEIT = "EUROPAWEIT"  #: europaweit



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.tagesvektor import Tagesvektor
from bo4e.com.zeitreihenwertkompakt import Zeitreihenwertkompakt
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_sigmoidparameter import example_sigmoidparameter

example_tagesvektor: Tagesvektor = Tagesvektor(
    tag=datetime(2021, 12, 15, 5, 0, tzinfo=timezone.utc),
    werte=[
        Zeitreihenwertkompakt(
            wert=Decimal(40),
        ),
        Zeitreihenwertkompakt(
            wert=Decimal(50),
        ),
    ],
)
example_tagesvektor_json = {
    "tag": datetime(2021, 12, 15, 5, 0, tzinfo=timezone.utc),
    "werte": [
        {"wert": Decimal("40"), "statuszusatz": None, "status": None},
        {"wert": Decimal("50"), "statuszusatz": None, "status": None},
    ],
}


class TestTagesvektor:
    @pytest.mark.parametrize(
        "tagesvektor, expected_json_dict",
        [
            pytest.param(
                example_tagesvektor,
                example_tagesvektor_json,
            ),
        ],
    )
    def test_serialization_roundtrip(self, tagesvektor: Tagesvektor, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of Preisstaffel.
        """
        assert_serialization_roundtrip(tagesvektor, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tagesvektor()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    def test_list_not_long_enough_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tagesvektor(tag=datetime(2021, 12, 15, 5, 0, tzinfo=timezone.utc), werte=[])

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)

    # add tests for issues 261 and 262 here



"""
Contains Tarif class and corresponding marshmallow schema for de-/serialization
"""

from datetime import datetime
from typing import List, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.tarifinfo import Tarifinfo
from bo4e.com.aufabschlagregional import AufAbschlagRegional
from bo4e.com.preisgarantie import Preisgarantie
from bo4e.com.tarifberechnungsparameter import Tarifberechnungsparameter
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.com.tarifpreispositionproort import TarifpreispositionProOrt
from bo4e.enum.botyp import BoTyp


class Tarif(Tarifinfo):
    """
    Abbildung eines Tarifs mit regionaler Zuordnung von Preisen und Auf- und Abschlägen

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Tarif.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarif JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Tarif.json>`_

    """

    bo_typ: BoTyp = BoTyp.TARIF
    # required attributes
    #: Gibt an, wann der Preis zuletzt angepasst wurde
    preisstand: datetime
    #: Für die Berechnung der Kosten sind die hier abgebildeten Parameter heranzuziehen
    berechnungsparameter: Tarifberechnungsparameter
    #: Die festgelegten Preise mit regionaler Eingrenzung z.B. für Arbeitspreis, Grundpreis etc.
    tarifpreise: conlist(TarifpreispositionProOrt, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Auf- und Abschläge auf die Preise oder Kosten mit regionaler Eingrenzung
    tarif_auf_abschlaege: Optional[List[AufAbschlagRegional]] = None
    # todo: fix inconsistency: RegionalerAufAbschlag vs. AufAbschlagRegional
    # https://github.com/Hochfrequenz/BO4E-python/issues/345

    #: Preisgarantie für diesen Tarif
    preisgarantie: Optional[Preisgarantie] = None
    # todo: fix inconsistency with regionaltarif https://github.com/Hochfrequenz/BO4E-python/issues/346
    #: Die Bedingungen und Einschränkungen unter denen ein Tarif angewendet werden kann
    tarifeinschraenkung: Optional[Tarifeinschraenkung] = None



"""
Contains Zeitreihenwert class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import Any, Dict

from pydantic import validator

from bo4e.com.zeitreihenwertkompakt import Zeitreihenwertkompakt
from bo4e.validators import check_bis_is_later_than_von

# pylint: disable=too-few-public-methods


class Zeitreihenwert(Zeitreihenwertkompakt):
    """
    Abbildung eines Zeitreihenwertes bestehend aus Zeitraum, Wert und Statusinformationen.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zeitreihenwert.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zeitreihenwert JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Zeitreihenwert.json>`_

    """

    # required attributes
    datum_uhrzeit_von: datetime  #: Datum Uhrzeit mit Auflösung Sekunden an dem das Messintervall begonnen wurde (inklusiv)
    datum_uhrzeit_bis: datetime  #: Datum Uhrzeit mit Auflösung Sekunden an dem das Messintervall endet (exklusiv)
    _bis_check = validator("datum_uhrzeit_bis", allow_reuse=True)(check_bis_is_later_than_von)

    @staticmethod
    def _get_inclusive_start(values: Dict[str, Any]) -> datetime:
        """return the inclusive start (used in the validator)"""
        return values["datum_uhrzeit_von"]

    # def _get_exclusive_end(self) -> datetime:
    #     """return the exclusive end (used in the validator)"""
    #     return self.datum_uhrzeit_bis



import pytest
from pydantic import ValidationError

from bo4e.bo.fremdkosten import Fremdkosten
from bo4e.com.fremdkostenblock import Fremdkostenblock
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_betrag import example_betrag
from tests.test_zeitraum import example_zeitraum


class TestFremdkosten:
    @pytest.mark.parametrize(
        "fremdkosten",
        [
            pytest.param(
                Fremdkosten(
                    gueltigkeit=example_zeitraum,
                    summe_kosten=example_betrag,
                    kostenbloecke=[Fremdkostenblock(kostenblockbezeichnung="teststring")],
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, fremdkosten: Fremdkosten) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(fremdkosten)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Fremdkosten()  # type: ignore[call-arg]
        assert "1 validation error" in str(excinfo.value)



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.aufabschlagproort import AufAbschlagProOrt
from bo4e.com.aufabschlagstaffelproort import AufAbschlagstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip


class TestAufAbschlagProOrt:
    @pytest.mark.parametrize(
        "aufabschlagproort, expected_json_dict",
        [
            pytest.param(
                AufAbschlagProOrt(
                    postleitzahl="01187",
                    ort="Dresden",
                    netznr="2",
                    staffeln=[
                        AufAbschlagstaffelProOrt(
                            wert=Decimal(2.5),
                            staffelgrenze_von=Decimal(1),
                            staffelgrenze_bis=Decimal(5),
                        )
                    ],
                ),
                {
                    "postleitzahl": "01187",
                    "ort": "Dresden",
                    "netznr": "2",
                    "staffeln": [
                        {
                            "wert": Decimal("2.5"),
                            "staffelgrenzeVon": Decimal("1"),
                            "staffelgrenzeBis": Decimal("5"),
                        }
                    ],
                },
            )
        ],
    )
    def test_serialization_roundtrip(
        self, aufabschlagproort: AufAbschlagProOrt, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of AufAbschlagProOrt with minimal attributes.
        """
        assert_serialization_roundtrip(aufabschlagproort, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = AufAbschlagProOrt()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)

    def test_failing_validation_list_length_at_least_one(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = AufAbschlagProOrt(
                postleitzahl="01187",
                ort="Dresden",
                netznr="2",
                staffeln=[],
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



import pytest
from pydantic import ValidationError

from bo4e.bo.energiemenge import Energiemenge
from bo4e.enum.lokationstyp import Lokationstyp
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_verbrauch import example_verbrauch


class TestEnergiemenge:
    @pytest.mark.parametrize(
        "energiemenge",
        [
            pytest.param(
                Energiemenge(
                    lokations_id="DE0123456789012345678901234567890",
                    lokationstyp=Lokationstyp.MELO,
                    energieverbrauch=[example_verbrauch],
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, energiemenge: Energiemenge) -> None:
        assert_serialization_roundtrip(energiemenge)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Energiemenge()  # type: ignore[call-arg]

        assert "3 validation errors" in str(excinfo.value)



from datetime import datetime, timezone
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.kriteriumwert import KriteriumWert
from bo4e.com.regionalegueltigkeit import RegionaleGueltigkeit
from bo4e.com.regionalepreisgarantie import RegionalePreisgarantie
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp
from bo4e.enum.preisgarantietyp import Preisgarantietyp
from bo4e.enum.tarifregionskriterium import Tarifregionskriterium
from tests.serialization_helper import assert_serialization_roundtrip

example_regionale_preisgarantie = RegionalePreisgarantie(
    preisgarantietyp=Preisgarantietyp.NUR_ENERGIEPREIS,
    zeitliche_gueltigkeit=Zeitraum(
        startzeitpunkt=datetime(2011, 2, 5, 16, 43, tzinfo=timezone.utc),
        endzeitpunkt=datetime(2021, 7, 30, tzinfo=timezone.utc),
    ),
    regionale_gueltigkeit=RegionaleGueltigkeit(
        gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
        kriteriums_werte=[KriteriumWert(kriterium=Tarifregionskriterium.POSTLEITZAHL, wert="01069")],
    ),
)


class TestRegionalePreisgarantie:
    @pytest.mark.parametrize(
        "regionale_preisgarantie, expected_json_dict",
        [
            pytest.param(
                example_regionale_preisgarantie,
                {
                    "beschreibung": None,
                    "preisgarantietyp": "NUR_ENERGIEPREIS",
                    "regionaleGueltigkeit": {
                        "gueltigkeitstyp": "NUR_IN",
                        "kriteriumsWerte": [{"kriterium": Tarifregionskriterium.POSTLEITZAHL, "wert": "01069"}],
                    },
                    "zeitlicheGueltigkeit": {
                        "startdatum": None,
                        "einheit": None,
                        "enddatum": None,
                        "dauer": None,
                        "endzeitpunkt": datetime(2021, 7, 30, 0, 0, tzinfo=timezone.utc),
                        "startzeitpunkt": datetime(2011, 2, 5, 16, 43, tzinfo=timezone.utc),
                    },
                },
                id="only required attributes",
            ),
        ],
    )
    def test_regionale_preisgarantie_serialization_roundtrip(
        self, regionale_preisgarantie: RegionalePreisgarantie, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of RegionalePreisgarantie with minimal attributes.
        """
        assert_serialization_roundtrip(regionale_preisgarantie, expected_json_dict)

    def test_regionalepreisgarantie_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = RegionalePreisgarantie()  # type: ignore[call-arg]

        assert "3 validation errors" in str(excinfo.value)



"""
Contains Marktteilnehmer class
and corresponding marshmallow schema for de-/serialization
"""


# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import constr

from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.enum.botyp import BoTyp
from bo4e.enum.marktrolle import Marktrolle
from bo4e.enum.rollencodetyp import Rollencodetyp
from bo4e.enum.sparte import Sparte


class Marktteilnehmer(Geschaeftspartner):
    """
    Objekt zur Aufnahme der Information zu einem Marktteilnehmer

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Marktteilnehmer.svg" type="image/svg+xml"></object>

    .. HINT::
        `Marktteilnehmer JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Marktteilnehmer.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.MARKTTEILNEHMER
    #: Gibt im Klartext die Bezeichnung der Marktrolle an
    marktrolle: Marktrolle
    #: Gibt die Codenummer der Marktrolle an
    rollencodenummer: constr(strict=True, regex=r"^\d{13}$")  # type: ignore[valid-type]
    #: Gibt den Typ des Codes an
    rollencodetyp: Rollencodetyp
    #: Sparte des Marktteilnehmers, z.B. Gas oder Strom
    sparte: Sparte

    # optional attributes
    #: Die 1:1-Kommunikationsadresse des Marktteilnehmers; Diese wird in der Marktkommunikation verwendet.
    makoadresse: Optional[str] = None



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Preisgarantietyp(StrEnum):
    """
    Aufzählung der Möglichkeiten für die Vergabe von Preisgarantien
    """

    ALLE_PREISBESTANDTEILE_BRUTTO = "ALLE_PREISBESTANDTEILE_BRUTTO"  #: ALLE_PREISBESTANDTEILE_BRUTTO
    ALLE_PREISBESTANDTEILE_NETTO = "ALLE_PREISBESTANDTEILE_NETTO"  #: ALLE_PREISBESTANDTEILE_NETTO
    PREISBESTANDTEILE_OHNE_ABGABEN = "PREISBESTANDTEILE_OHNE_ABGABEN"  #: PREISBESTANDTEILE_OHNE_ABGABEN
    NUR_ENERGIEPREIS = "NUR_ENERGIEPREIS"  #: NUR_ENERGIEPREIS



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Gueltigkeitstyp(StrEnum):
    """
    Übersicht der verschiedenen Gültigkeiten zur Umsetzung von Positiv- bzw. Negativlisten.
    """

    NUR_IN = "NUR_IN"  #: Ein so eingeschränktes Merkmal gilt nur mit den angebenen Werten
    NICHT_IN = "NICHT_IN"  #: Ein so eingeschränktes Merkmal gilt nicht mit den angebenen Werten
    NUR_IN_KOMBINATION_MIT = (
        "NUR_IN_KOMBINATION_MIT"  #: Die Kriterien mit diesem Gültigkeitstyp werden miteinander kombiniert
    )



from datetime import datetime, timezone
from typing import Any, Dict, List

import pytest
from pydantic import ValidationError

from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.bo.vertrag import Vertrag
from bo4e.com.adresse import Adresse
from bo4e.com.unterschrift import Unterschrift
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.com.vertragsteil import Vertragsteil
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart
from bo4e.enum.landescode import Landescode
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsart import Vertragsart
from bo4e.enum.vertragsstatus import Vertragsstatus


class TestVertrag:
    _vertragsnummer = "1234567891011"
    _vertragsart = Vertragsart.BILANZIERUNGSVERTRAG
    _vertragsstatus = Vertragsstatus.AKTIV
    _sparte = Sparte.STROM
    _vertragsbeginn = datetime(2021, 4, 30, 13, 45, tzinfo=timezone.utc)
    _vertragsende = datetime(2021, 6, 5, 16, 30, tzinfo=timezone.utc)
    _vertragspartner1 = Geschaeftspartner(
        anrede=Anrede.FRAU,
        name1="von Sinnen",
        name2="Helga",
        name3=None,
        gewerbekennzeichnung=True,
        kontaktweg=[Kontaktart.E_MAIL],
        umsatzsteuer_id="DE267311963",
        glaeubiger_id="DE98ZZZ09999999999",
        e_mail_adresse="test@bo4e.de",
        website="bo4e.de",
        geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
        partneradresse=Adresse(
            postleitzahl="24306",
            ort="Plön",
            strasse="Kirchstraße",
            hausnummer="3",
        ),
    )
    _vertragspartner2 = Geschaeftspartner(
        name1="Eckart",
        name2="Björn",
        gewerbekennzeichnung=False,
        geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
        partneradresse=Adresse(
            postleitzahl="24211",
            ort="Preetz",
            strasse="Am Markt",
            hausnummer="67",
        ),
    )
    _vertragsteile = [
        Vertragsteil(
            vertragsteilbeginn=datetime(2021, 4, 30, tzinfo=timezone.utc),
            vertragsteilende=datetime(2021, 6, 5, tzinfo=timezone.utc),
        )
    ]
    _vertragspartner1_dict: Dict[str, Any] = {
        "versionstruktur": "2",
        "boTyp": BoTyp.GESCHAEFTSPARTNER,
        "externeReferenzen": [],
        "name1": "von Sinnen",
        "gewerbekennzeichnung": True,
        "geschaeftspartnerrolle": [Geschaeftspartnerrolle.DIENSTLEISTER],
        "anrede": "FRAU",
        "name2": "Helga",
        "name3": None,
        "hrnummer": None,
        "amtsgericht": None,
        "kontaktweg": [Kontaktart.E_MAIL],
        "umsatzsteuerId": "DE267311963",
        "glaeubigerId": "DE98ZZZ09999999999",
        "eMailAdresse": "test@bo4e.de",
        "website": "bo4e.de",
        "partneradresse": {
            "postleitzahl": "24306",
            "ort": "Plön",
            "strasse": "Kirchstraße",
            "hausnummer": "3",
            "postfach": None,
            "adresszusatz": None,
            "coErgaenzung": None,
            "landescode": Landescode.DE,  # type:ignore[attr-defined]
        },
    }
    _vertragspartner2_dict: Dict[str, Any] = {
        "versionstruktur": "2",
        "boTyp": BoTyp.GESCHAEFTSPARTNER,
        "externeReferenzen": [],
        "name1": "Eckart",
        "gewerbekennzeichnung": False,
        "geschaeftspartnerrolle": [Geschaeftspartnerrolle.DIENSTLEISTER],
        "anrede": None,
        "name2": "Björn",
        "name3": None,
        "hrnummer": None,
        "amtsgericht": None,
        "kontaktweg": [],
        "umsatzsteuerId": None,
        "glaeubigerId": None,
        "eMailAdresse": None,
        "website": None,
        "partneradresse": {
            "postleitzahl": "24211",
            "ort": "Preetz",
            "strasse": "Am Markt",
            "hausnummer": "67",
            "postfach": None,
            "adresszusatz": None,
            "coErgaenzung": None,
            "landescode": Landescode.DE,  # type:ignore[attr-defined]
        },
    }
    _vertragsteile_dict: List[Dict[str, Any]] = [
        {
            "vertragsteilbeginn": datetime(2021, 4, 30, tzinfo=timezone.utc),
            "vertragsteilende": datetime(2021, 6, 5, tzinfo=timezone.utc),
            "lokation": None,
            "vertraglichFixierteMenge": None,
            "minimaleAbnahmemenge": None,
            "maximaleAbnahmemenge": None,
        }
    ]

    def get_example_vertrag(self) -> Vertrag:
        return Vertrag(
            vertragsnummer=self._vertragsnummer,
            vertragsart=self._vertragsart,
            vertragsstatus=self._vertragsstatus,
            sparte=self._sparte,
            vertragsbeginn=self._vertragsbeginn,
            vertragsende=self._vertragsende,
            vertragspartner1=self._vertragspartner1,
            vertragspartner2=self._vertragspartner2,
            vertragsteile=self._vertragsteile,
        )

    def get_example_vertrag_expected_dict(self) -> Dict[str, Any]:
        return {
            "vertragsnummer": self._vertragsnummer,
            "vertragsart": self._vertragsart,
            "vertragsstatus": self._vertragsstatus,
            "sparte": self._sparte,
            "vertragsbeginn": self._vertragsbeginn,
            "vertragsende": self._vertragsende,
            "vertragspartner1": self._vertragspartner1_dict,
            "vertragspartner2": self._vertragspartner2_dict,
            "vertragsteile": self._vertragsteile_dict,
            "versionstruktur": "2",
            "boTyp": BoTyp.VERTRAG,
            "externeReferenzen": [],
            "beschreibung": None,
            "vertragskonditionen": None,
            "unterzeichnervp1": None,
            "unterzeichnervp2": None,
        }

    def test_serialisation_only_required_attributes(self) -> None:
        """
        Test de-/serialisation of Vertrag with minimal attributes.
        """
        vertrag = self.get_example_vertrag()

        json_string = vertrag.json(by_alias=True, ensure_ascii=False)

        assert vertrag.bo_typ is BoTyp.VERTRAG, "boTyp was not automatically set"
        assert self._vertragsnummer in json_string
        assert "BILANZIERUNGSVERTRAG" in json_string
        assert "AKTIV" in json_string
        assert "STROM" in json_string
        assert "2021-04-30T13:45:00+00:00" in json_string
        assert "2021-06-05T16:30:00+00:00" in json_string
        assert "von Sinnen" in json_string
        assert "Preetz" in json_string
        assert "2021-06-05T00:00:00+00:00" in json_string

        vertrag_deserialized = Vertrag.parse_raw(json_string)

        assert vertrag_deserialized.vertragsnummer == self._vertragsnummer
        assert vertrag_deserialized.vertragsart == self._vertragsart
        assert vertrag_deserialized.vertragsstatus == self._vertragsstatus
        assert vertrag_deserialized.sparte == self._sparte
        assert vertrag_deserialized.vertragsbeginn == self._vertragsbeginn
        assert vertrag_deserialized.vertragsende == self._vertragsende
        assert vertrag_deserialized.vertragspartner1 == self._vertragspartner1
        assert vertrag_deserialized.vertragspartner2 == self._vertragspartner2
        assert vertrag_deserialized.vertragsteile == self._vertragsteile

    def test_serialisation_required_and_optional_attributes(self) -> None:
        """
        Test de-/serialisation of Vertrag with maximal attributes.
        """

        vertrag = Vertrag(
            vertragsnummer=self._vertragsnummer,
            vertragsart=self._vertragsart,
            vertragsstatus=self._vertragsstatus,
            sparte=self._sparte,
            vertragsbeginn=self._vertragsbeginn,
            vertragsende=self._vertragsende,
            vertragspartner1=self._vertragspartner1,
            vertragspartner2=self._vertragspartner2,
            vertragsteile=[
                Vertragsteil(
                    vertragsteilbeginn=datetime(2021, 4, 30, tzinfo=timezone.utc),
                    vertragsteilende=datetime(2021, 6, 5, tzinfo=timezone.utc),
                ),
                Vertragsteil(
                    vertragsteilbeginn=datetime(2001, 1, 23, tzinfo=timezone.utc),
                    vertragsteilende=datetime(2002, 12, 3, tzinfo=timezone.utc),
                ),
            ],
            beschreibung="Hello Vertrag",
            vertragskonditionen=Vertragskonditionen(beschreibung="Beschreibung"),
            unterzeichnervp1=[Unterschrift(name="Foo")],
            unterzeichnervp2=[Unterschrift(name="Bar"), Unterschrift(name="Dr.No")],
        )

        json_string = vertrag.json(by_alias=True, ensure_ascii=False)

        assert vertrag.bo_typ is BoTyp.VERTRAG, "boTyp was not automatically set"
        assert self._vertragsnummer in json_string
        assert "BILANZIERUNGSVERTRAG" in json_string
        assert "AKTIV" in json_string
        assert "STROM" in json_string
        assert "2021-04-30T13:45:00+00:00" in json_string
        assert "2021-06-05T16:30:00+00:00" in json_string
        assert "von Sinnen" in json_string
        assert "Preetz" in json_string
        assert "2021-06-05T00:00:00+00:00" in json_string
        assert "2002-12-03T00:00:00+00:00" in json_string
        assert "Hello Vertrag" in json_string
        assert "Beschreibung" in json_string
        assert "Foo" in json_string
        assert "Bar" in json_string
        assert "Dr.No" in json_string

        vertrag_deserialized = Vertrag.parse_raw(json_string)

        assert vertrag_deserialized.vertragsnummer == self._vertragsnummer
        assert vertrag_deserialized.vertragsart == self._vertragsart
        assert vertrag_deserialized.vertragsstatus == self._vertragsstatus
        assert vertrag_deserialized.sparte == self._sparte
        assert vertrag_deserialized.vertragsbeginn == self._vertragsbeginn
        assert vertrag_deserialized.vertragsende == self._vertragsende
        assert vertrag_deserialized.vertragspartner1 == self._vertragspartner1
        assert vertrag_deserialized.vertragspartner2 == self._vertragspartner2
        assert vertrag_deserialized.vertragsteile == [
            Vertragsteil(
                vertragsteilbeginn=datetime(2021, 4, 30, tzinfo=timezone.utc),
                vertragsteilende=datetime(2021, 6, 5, tzinfo=timezone.utc),
            ),
            Vertragsteil(
                vertragsteilbeginn=datetime(2001, 1, 23, tzinfo=timezone.utc),
                vertragsteilende=datetime(2002, 12, 3, tzinfo=timezone.utc),
            ),
        ]
        assert vertrag_deserialized.beschreibung == "Hello Vertrag"
        assert vertrag_deserialized.vertragskonditionen == Vertragskonditionen(beschreibung="Beschreibung")
        assert vertrag_deserialized.unterzeichnervp1 == [Unterschrift(name="Foo")]
        assert vertrag_deserialized.unterzeichnervp2 == [Unterschrift(name="Bar"), Unterschrift(name="Dr.No")]

    def test_missing_required_attributes(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Vertrag()  # type: ignore[call-arg]

        assert "9 validation errors" in str(excinfo.value)

    def test_serialization_fails_for_empty_vertragsteile(self) -> None:
        """
        Test serialisation of Zaehler fails if there are no vertragsteile.
        """
        with pytest.raises(ValidationError) as excinfo:
            _ = Vertrag(
                vertragsnummer=self._vertragsnummer,
                vertragsart=self._vertragsart,
                vertragsstatus=self._vertragsstatus,
                sparte=self._sparte,
                vertragsbeginn=self._vertragsbeginn,
                vertragsende=self._vertragsende,
                vertragspartner1=self._vertragspartner1,
                vertragspartner2=self._vertragspartner2,
                vertragsteile=[],
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



"""
Contains Standorteigenschaften class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.standorteigenschaftengas import StandorteigenschaftenGas
from bo4e.com.standorteigenschaftenstrom import StandorteigenschaftenStrom
from bo4e.enum.botyp import BoTyp


class Standorteigenschaften(Geschaeftsobjekt):
    """
    Modelliert die regionalen und spartenspezifischen Eigenschaften einer gegebenen Adresse.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Standorteigenschaften.svg" type="image/svg+xml"></object>

    .. HINT::
        `Standorteigenschaften JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Standorteigenschaften.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.STANDORTEIGENSCHAFTEN
    #: Eigenschaften zur Sparte Strom
    eigenschaften_strom: conlist(StandorteigenschaftenStrom, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Eigenschaften zur Sparte Gas
    eigenschaften_gas: Optional[StandorteigenschaftenGas] = None



"""
Contains Vertragskonditionen class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal
from typing import Optional

from bo4e.com.com import COM
from bo4e.com.zeitraum import Zeitraum

# pylint: disable=too-few-public-methods


class Vertragskonditionen(COM):
    """
    Abbildung für Vertragskonditionen. Die Komponente wird sowohl im Vertrag als auch im Tarif verwendet.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Vertragskonditionen.svg" type="image/svg+xml"></object>

    .. HINT::
        `Vertragskonditionen JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Vertragskonditionen.json>`_

    """

    # optional attributes
    #: Freitext zur Beschreibung der Konditionen, z.B. "Standardkonditionen Gas"
    beschreibung: Optional[str] = None
    #: Anzahl der vereinbarten Abschläge pro Jahr, z.B. 12
    anzahl_abschlaege: Optional[Decimal] = None
    #: Über diesen Zeitraum läuft der Vertrag
    vertragslaufzeit: Optional[Zeitraum] = None
    #: Innerhalb dieser Frist kann der Vertrag gekündigt werden
    kuendigungsfrist: Optional[Zeitraum] = None
    #: Falls der Vertrag nicht gekündigt wird, verlängert er sich automatisch um die hier angegebene Zeit
    vertragsverlaengerung: Optional[Zeitraum] = None
    #: In diesen Zyklen werden Abschläge gestellt. Alternativ kann auch die Anzahl in den Konditionen angeben werden.
    abschlagszyklus: Optional[Zeitraum] = None



"""
Contains Tarifpreisblatt class and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import List, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.bo.tarifinfo import Tarifinfo
from bo4e.com.aufabschlag import AufAbschlag
from bo4e.com.preisgarantie import Preisgarantie
from bo4e.com.tarifberechnungsparameter import Tarifberechnungsparameter
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.com.tarifpreisposition import Tarifpreisposition
from bo4e.enum.botyp import BoTyp


class Tarifpreisblatt(Tarifinfo):
    """
    Tarifinformation mit Preisen, Aufschlägen und Berechnungssystematik

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Tarifpreisblatt.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifpreisblatt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Tarifpreisblatt.json>`_

    """

    bo_typ: BoTyp = BoTyp.TARIFPREISBLATT
    # required attributes (additional to those of Tarifinfo)
    #: Gibt an, wann der Preis zuletzt angepasst wurde
    preisstand: datetime
    #: Die festgelegten Preise, z.B. für Arbeitspreis, Grundpreis etc.
    tarifpreise: conlist(Tarifpreisposition, min_items=1)  # type: ignore[valid-type]
    #: Für die Berechnung der Kosten sind die hier abgebildeten Parameter heranzuziehen
    berechnungsparameter: Tarifberechnungsparameter

    # optional attributes
    #: Die Bedingungen und Einschränkungen unter denen ein Tarif angewendet werden kann
    tarifeinschraenkung: Optional[Tarifeinschraenkung] = None
    #: Festlegung von Garantien für bestimmte Preisanteile
    preisgarantie: Optional[Preisgarantie] = None
    #: Auf- und Abschläge auf die Preise oder Kosten
    tarif_auf_abschlaege: Optional[List[AufAbschlag]] = None



"""
Contains Betrag class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal

from bo4e.com.com import COM
from bo4e.enum.waehrungscode import Waehrungscode

# pylint: disable=too-few-public-methods


class Betrag(COM):
    """
    Die Komponente wird dazu verwendet, Summenbeträge (beispielsweise in Angeboten und Rechnungen) als Geldbeträge
    abzubilden. Die Einheit ist dabei immer die Hauptwährung also Euro, Dollar etc…

    .. raw:: html

        <object data="../_static/images/bo4e/com/Betrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Betrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Betrag.json>`_

    """

    # required attributes
    wert: Decimal  #: Gibt den Betrag des Preises an.
    waehrung: Waehrungscode  #: Die entsprechende Waehrung



"""
Contains Menge class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal

from bo4e.com.com import COM
from bo4e.enum.mengeneinheit import Mengeneinheit

# pylint: disable=too-few-public-methods


class Menge(COM):
    """
    Abbildung einer Menge mit Wert und Einheit.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Menge.svg" type="image/svg+xml"></object>

    .. HINT::
        `Menge JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Menge.json>`_

    """

    # required attributes
    #: Gibt den absoluten Wert der Menge an
    wert: Decimal
    #: Gibt die Einheit zum jeweiligen Wert an
    einheit: Mengeneinheit



from datetime import datetime, timezone

import pytest
from pydantic import ValidationError

from bo4e.com.rechnungsposition import Rechnungsposition
from bo4e.enum.bdewartikelnummer import BDEWArtikelnummer
from bo4e.enum.zeiteinheit import Zeiteinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_betrag import example_betrag
from tests.test_menge import example_menge
from tests.test_preis import example_preis
from tests.test_steuerbetrag import example_steuerbetrag


class TestRechnungsposition:
    @pytest.mark.parametrize(
        "rechnungsposition",
        [
            pytest.param(
                Rechnungsposition(
                    positionsnummer=1,
                    lieferung_von=datetime(2021, 3, 15, tzinfo=timezone.utc),
                    lieferung_bis=datetime(2022, 3, 15, tzinfo=timezone.utc),
                    positionstext="Besonders wertvolle Rechnungsposition",
                    zeiteinheit=Zeiteinheit.JAHR,
                    artikelnummer=BDEWArtikelnummer.AUSGLEICHSENERGIE_UNTERDECKUNG,
                    lokations_id="51238696781",
                    positions_menge=example_menge,
                    zeitbezogene_menge=example_menge,
                    einzelpreis=example_preis,
                    teilsumme_netto=example_betrag,
                    teilrabatt_netto=example_betrag,
                    teilsumme_steuer=example_steuerbetrag,
                    artikel_id="7-8-9",
                ),
                id="maximal attributes",
            )
        ],
    )
    def test_serialization_roundtrip(self, rechnungsposition: Rechnungsposition) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(rechnungsposition)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Rechnungsposition()  # type: ignore[call-arg]
        assert "8 validation errors" in str(excinfo.value)



"""
Contains Geokoordinaten class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods


class Geokoordinaten(COM):
    """
    This component provides the geo-coordinates for a location.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Geokoordinaten.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geokoordinaten JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Geokoordinaten.json>`_

    """

    breitengrad: Decimal
    laengengrad: Decimal



"""
Contains Zeitreihe class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.zeitreihenwert import Zeitreihenwert
from bo4e.enum.botyp import BoTyp
from bo4e.enum.medium import Medium
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.messart import Messart
from bo4e.enum.messgroesse import Messgroesse
from bo4e.enum.wertermittlungsverfahren import Wertermittlungsverfahren


class Zeitreihe(Geschaeftsobjekt):
    """
    Abbildung einer allgemeinen Zeitreihe mit einem Wertvektor.
    Die Werte können mit wahlfreier zeitlicher Distanz im Vektor abgelegt sein.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Zeitreihe.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zeitreihe JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Zeitreihe.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.ZEITREIHE
    #: Bezeichnung für die Zeitreihe
    bezeichnung: str
    #: Beschreibt, was gemessen wurde (z.B. Strom, Spannung, Wirkleistung, Scheinleistung)
    messgroesse: Messgroesse
    #: Beschreibt die Art der Messung (z.B. aktueller Wert, mittlerer Wert, maximaler Wert)
    messart: Messart
    #: Medium, das gemessen wurde (z.B. Wasser, Dampf, Strom, Gas)
    medium: Medium
    #: Alle Werte in der Tabelle haben die Einheit, die hier angegeben ist
    einheit: Mengeneinheit
    #: Hier liegen jeweils die Werte
    werte: conlist(Zeitreihenwert, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Beschreibt die Verwendung der Zeitreihe
    beschreibung: Optional[str] = None
    #: Version der Zeitreihe
    version: Optional[str] = None
    #: Kennzeichnung, wie die Werte entstanden sind, z.B. durch Messung
    wertherkunft: Optional[Wertermittlungsverfahren] = None



import pytest
from pydantic import ValidationError

from bo4e.bo.kosten import Kosten
from bo4e.com.kostenblock import Kostenblock
from bo4e.enum.kostenklasse import Kostenklasse
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_betrag import example_betrag
from tests.test_zeitraum import example_zeitraum

example_kosten = Kosten(
    kostenklasse=Kostenklasse.FREMDKOSTEN,
    gueltigkeit=example_zeitraum,
    kostenbloecke=[
        Kostenblock(
            kostenblockbezeichnung="Mein Kostenblock",
        )
    ],
    summe_kosten=[example_betrag],
)


class TestKosten:
    @pytest.mark.parametrize(
        "kosten",
        [
            pytest.param(
                example_kosten,
                id="maximal attributes",
            )
        ],
    )
    def test_serialization_roundtrip(self, kosten: Kosten) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(kosten)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Kosten()  # type: ignore[call-arg]
        assert "3 validation errors" in str(excinfo.value)



from bo4e.bo.ansprechpartner import Ansprechpartner
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse
from bo4e.com.rufnummer import Rufnummer
from bo4e.com.zustaendigkeit import Zustaendigkeit
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart
from bo4e.enum.rufnummernart import Rufnummernart
from bo4e.enum.themengebiet import Themengebiet
from bo4e.enum.titel import Titel


class TestAnsprechpartner:
    def test_de_serialisation_minimal_attributes(self) -> None:
        """
        Test de-/serialisation of Ansprechpartner only with required attributes
        """
        ansprechpartner = Ansprechpartner(
            # required attributes
            nachname="Müller-Schmidt",
            geschaeftspartner=Geschaeftspartner(
                anrede=Anrede.FRAU,
                name1="von Sinnen",
                name2="Helga",
                name3=None,
                gewerbekennzeichnung=True,
                hrnummer="HRB 254466",
                amtsgericht="Amtsgericht München",
                kontaktweg=[Kontaktart.E_MAIL],
                umsatzsteuer_id="DE267311963",
                glaeubiger_id="DE98ZZZ09999999999",
                e_mail_adresse="test@bo4e.de",
                website="bo4e.de",
                geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
                partneradresse=Adresse(
                    postleitzahl="12345",
                    ort="Sankt Musterweiler",
                    strasse="Mühlenweg",
                    hausnummer="540a",
                ),
            ),
        )
        assert ansprechpartner.versionstruktur == "2", "versionstruktur was not automatically set"
        assert ansprechpartner.bo_typ is BoTyp.ANSPRECHPARTNER, "boTyp was not automatically set"

        json_string = ansprechpartner.json(by_alias=True, ensure_ascii=False)
        assert "Müller-Schmidt" in json_string
        assert "Mühlenweg" in json_string
        assert '"FRAU"' in json_string

        deserialized_ansprechpartner = Ansprechpartner.parse_raw(json_string)

        assert isinstance(deserialized_ansprechpartner, Ansprechpartner)
        assert isinstance(deserialized_ansprechpartner.geschaeftspartner, Geschaeftspartner)
        assert deserialized_ansprechpartner == ansprechpartner

    def test_de_serialisation_maximal_attributes(self) -> None:
        """
        Test de-/serialisation of Ansprechpartner only with required attributes
        """
        ansprechpartner = Ansprechpartner(
            # required attributes
            nachname="Müller-Schmidt",
            geschaeftspartner=Geschaeftspartner(
                anrede=Anrede.FRAU,
                name1="von Sinnen",
                name2="Helga",
                name3=None,
                gewerbekennzeichnung=True,
                hrnummer="HRB 254466",
                amtsgericht="Amtsgericht München",
                kontaktweg=[Kontaktart.E_MAIL],
                umsatzsteuer_id="DE267311963",
                glaeubiger_id="DE98ZZZ09999999999",
                e_mail_adresse="test@bo4e.de",
                website="bo4e.de",
                geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
                partneradresse=Adresse(
                    postleitzahl="12345",
                    ort="Sankt Musterweiler",
                    strasse="Mühlenweg",
                    hausnummer="540a",
                ),
            ),
            # optional attributes
            anrede=Anrede.EHELEUTE,
            individuelle_anrede="Künstler",
            titel=Titel.PROF_DR,
            vorname="Hans",
            e_mail_adresse="hans.müller@getrei.de",
            kommentar="does this thing work?",
            adresse=Adresse(
                postleitzahl="12345",
                ort="Sankt Musterweiler",
                strasse="Mühlenweg",
                hausnummer="540a",
            ),
            rufnummer=Rufnummer(nummerntyp=Rufnummernart.RUF_DURCHWAHL, rufnummer="112358132134"),
            zustaendigkeit=Zustaendigkeit(
                themengebiet=Themengebiet.MARKTKOMMUNIKATION, jobtitel="Schatzmeister", abteilung="unten rechts"
            ),
        )
        assert ansprechpartner.versionstruktur == "2", "versionstruktur was not automatically set"
        assert ansprechpartner.bo_typ is BoTyp.ANSPRECHPARTNER, "boTyp was not automatically set"

        json_string = ansprechpartner.json(by_alias=True, ensure_ascii=False)
        assert "Müller-Schmidt" in json_string
        assert "Mühlenweg" in json_string
        assert "PROF_DR" in json_string

        deserialized_ansprechpartner = Ansprechpartner.parse_raw(json_string)

        assert isinstance(deserialized_ansprechpartner, Ansprechpartner)
        assert isinstance(deserialized_ansprechpartner.geschaeftspartner, Geschaeftspartner)
        assert isinstance(deserialized_ansprechpartner.adresse, Adresse)
        assert isinstance(deserialized_ansprechpartner.rufnummer, Rufnummer)
        assert isinstance(deserialized_ansprechpartner.zustaendigkeit, Zustaendigkeit)
        assert deserialized_ansprechpartner == ansprechpartner



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Kostenklasse(StrEnum):
    """
    Kostenklassen bilden die oberste Ebene der verschiedenen Kosten.
    In der Regel werden die Gesamtkosten einer Kostenklasse in einer App berechnet.
    """

    FREMDKOSTEN = "FREMDKOSTEN"  #: Fremdkosten
    BESCHAFFUNG = "BESCHAFFUNG"  #: Beschaffung
    SELBSTKOSTEN = "SELBSTKOSTEN"  #: Selbstkosten
    MARGEN = "MARGEN"  #: Margen
    ENERGIEVERSORGUNGSKOSTEN = "ENERGIEVERSORGUNGSKOSTEN"  #: Energieversorgungskosten



"""
Contains Tarifpreis class
and corresponding marshmallow schema for de-/serialization
"""
from typing import Optional

from bo4e.com.preis import Preis
from bo4e.enum.preistyp import Preistyp

# pylint: disable=too-few-public-methods


class Tarifpreis(Preis):
    """
    Abbildung eines Tarifpreises mit Preistyp und Beschreibung abgeleitet von COM Preis.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Tarifpreis.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifpreis JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Tarifpreis.json>`_

    """

    # required attributes
    #:  Angabe des Preistypes (z.B. Grundpreis)
    preistyp: Preistyp

    # optional attributes
    #:  Beschreibung des Preises. Hier können z.B. Preisdetails angegeben sein, beispielsweise "Drehstromzähler".
    beschreibung: Optional[str] = None



"""
Contains Tarifkosten class
and corresponding marshmallow schema for de-/serialization
"""


from bo4e.bo.kosten import Kosten
from bo4e.bo.tarifinfo import Tarifinfo
from bo4e.enum.botyp import BoTyp

# pylint: disable=too-few-public-methods


class Tarifkosten(Tarifinfo):
    """
    Objekt zur Kommunikation von Kosten, die im Rahmen der Tarifanwendung entstehen

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Tarifkosten.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifkosten JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Tarifkosten.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.TARIFKOSTEN
    kosten: Kosten
    """
    Referenz (Link) zu einem Kostenobjekt, in dem die Kosten für die Anwendung
    des Tarifs auf eine Abnahmesituation berechnet wurden
    """

    # there are no optional attributes



"""
Contains RegionalePreisstaffel class and corresponding marshmallow schema for de-/serialization
"""


from bo4e.com.preisstaffel import Preisstaffel
from bo4e.com.regionalegueltigkeit import RegionaleGueltigkeit

# pylint: disable=too-few-public-methods


class RegionalePreisstaffel(Preisstaffel):
    """
    Abbildung einer Preisstaffel mit regionaler Abgrenzung

    .. raw:: html

        <object data="../_static/images/bo4e/com/RegionalePreisstaffel.svg" type="image/svg+xml"></object>

    .. HINT::
        `RegionalePreisstaffel JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/RegionalePreisstaffel.json>`_

    """

    # required attributes
    #: Regionale Eingrenzung der Preisstaffel
    regionale_gueltigkeit: RegionaleGueltigkeit



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.aufabschlagproort import AufAbschlagProOrt
from bo4e.com.aufabschlagregional import AufAbschlagRegional
from bo4e.com.aufabschlagstaffelproort import AufAbschlagstaffelProOrt
from bo4e.com.energieherkunft import Energieherkunft
from bo4e.com.energiemix import Energiemix
from bo4e.com.preisgarantie import Preisgarantie
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.aufabschlagsziel import AufAbschlagsziel
from bo4e.enum.erzeugungsart import Erzeugungsart
from bo4e.enum.preisgarantietyp import Preisgarantietyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip

example_aufabschlagregional = AufAbschlagRegional(
    bezeichnung="foo",
    betraege=[
        AufAbschlagProOrt(
            postleitzahl="01187",
            ort="Dresden",
            netznr="2",
            staffeln=[
                AufAbschlagstaffelProOrt(
                    wert=Decimal(2.5),
                    staffelgrenze_von=Decimal(1),
                    staffelgrenze_bis=Decimal(5),
                )
            ],
        ),
    ],
)


class TestAufAbschlagRegional:
    @pytest.mark.parametrize(
        "aufabschlagregional, expected_json_dict",
        [
            pytest.param(
                example_aufabschlagregional,
                {
                    "bezeichnung": "foo",
                    "betraege": [
                        {
                            "postleitzahl": "01187",
                            "ort": "Dresden",
                            "netznr": "2",
                            "staffeln": [
                                {
                                    "wert": Decimal("2.5"),
                                    "staffelgrenzeVon": Decimal("1"),
                                    "staffelgrenzeBis": Decimal("5"),
                                }
                            ],
                        },
                    ],
                    "beschreibung": None,
                    "aufAbschlagstyp": None,
                    "aufAbschlagsziel": None,
                    "einheit": None,
                    "website": None,
                    "zusatzprodukte": None,
                    "voraussetzungen": None,
                    "tarifnamensaenderungen": None,
                    "gueltigkeitszeitraum": None,
                    "energiemixaenderung": None,
                    "vertagskonditionsaenderung": None,
                    "garantieaenderung": None,
                    "einschraenkungsaenderung": None,
                },
                id="only required attributes",
            ),
            pytest.param(
                AufAbschlagRegional(
                    bezeichnung="foo",
                    betraege=[
                        AufAbschlagProOrt(
                            postleitzahl="01187",
                            ort="Dresden",
                            netznr="2",
                            staffeln=[
                                AufAbschlagstaffelProOrt(
                                    wert=Decimal(2.5),
                                    staffelgrenze_von=Decimal(1),
                                    staffelgrenze_bis=Decimal(5),
                                )
                            ],
                        ),
                    ],
                    beschreibung="bar",
                    auf_abschlagstyp=AufAbschlagstyp.RELATIV,
                    auf_abschlagsziel=AufAbschlagsziel.ARBEITSPREIS_HT,
                    einheit=Waehrungseinheit.EUR,
                    website="foo.bar",
                    zusatzprodukte=["Asterix", "Obelix"],
                    voraussetzungen=["Petterson", "Findus"],
                    tarifnamensaenderungen="foobar",
                    gueltigkeitszeitraum=Zeitraum(
                        startdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
                        enddatum=datetime(2020, 4, 1, tzinfo=timezone.utc),
                    ),
                    energiemixaenderung=Energiemix(
                        energiemixnummer=2,
                        energieart=Sparte.STROM,
                        bezeichnung="foo",
                        gueltigkeitsjahr=2021,
                        anteil=[
                            Energieherkunft(
                                erzeugungsart=Erzeugungsart.BIOGAS,
                                anteil_prozent=Decimal(40),
                            ),
                        ],
                    ),
                    vertagskonditionsaenderung=Vertragskonditionen(),
                    garantieaenderung=Preisgarantie(
                        preisgarantietyp=Preisgarantietyp.ALLE_PREISBESTANDTEILE_BRUTTO,
                        zeitliche_gueltigkeit=Zeitraum(
                            startdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
                            enddatum=datetime(2020, 4, 1, tzinfo=timezone.utc),
                        ),
                    ),
                    einschraenkungsaenderung=Tarifeinschraenkung(),
                ),
                {
                    "bezeichnung": "foo",
                    "betraege": [
                        {
                            "postleitzahl": "01187",
                            "ort": "Dresden",
                            "netznr": "2",
                            "staffeln": [
                                {
                                    "wert": Decimal("2.5"),
                                    "staffelgrenzeVon": Decimal("1"),
                                    "staffelgrenzeBis": Decimal("5"),
                                }
                            ],
                        },
                    ],
                    "beschreibung": "bar",
                    "aufAbschlagstyp": AufAbschlagstyp.RELATIV,
                    "aufAbschlagsziel": "ARBEITSPREIS_HT",
                    "einheit": Waehrungseinheit.EUR,
                    "website": "foo.bar",
                    "zusatzprodukte": ["Asterix", "Obelix"],
                    "voraussetzungen": ["Petterson", "Findus"],
                    "tarifnamensaenderungen": "foobar",
                    "gueltigkeitszeitraum": {
                        "startdatum": datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
                        "endzeitpunkt": None,
                        "einheit": None,
                        "enddatum": datetime(2020, 4, 1, 0, 0, tzinfo=timezone.utc),
                        "startzeitpunkt": None,
                        "dauer": None,
                    },
                    "energiemixaenderung": {
                        "energiemixnummer": 2,
                        "energieart": Sparte.STROM,
                        "bezeichnung": "foo",
                        "gueltigkeitsjahr": 2021,
                        "anteil": [
                            {
                                "erzeugungsart": Erzeugungsart.BIOGAS,
                                "anteilProzent": Decimal("40"),
                            }
                        ],
                        "oekolabel": [],
                        "bemerkung": None,
                        "co2Emission": None,
                        "atommuell": None,
                        "website": None,
                        "oekozertifikate": [],
                        "oekoTopTen": None,
                    },
                    "vertagskonditionsaenderung": {
                        "beschreibung": None,
                        "anzahlAbschlaege": None,
                        "vertragslaufzeit": None,
                        "kuendigungsfrist": None,
                        "vertragsverlaengerung": None,
                        "abschlagszyklus": None,
                    },
                    "garantieaenderung": {
                        "beschreibung": None,
                        "preisgarantietyp": "ALLE_PREISBESTANDTEILE_BRUTTO",
                        "zeitlicheGueltigkeit": {
                            "startdatum": datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
                            "endzeitpunkt": None,
                            "einheit": None,
                            "enddatum": datetime(2020, 4, 1, 0, 0, tzinfo=timezone.utc),
                            "startzeitpunkt": None,
                            "dauer": None,
                        },
                    },
                    "einschraenkungsaenderung": {
                        "zusatzprodukte": None,
                        "voraussetzungen": None,
                        "einschraenkungzaehler": None,
                        "einschraenkungleistung": None,
                    },
                },
                id="required and optional attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, aufabschlagregional: AufAbschlagRegional, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of AufAbschlagRegional with minimal attributes.
        """
        assert_serialization_roundtrip(aufabschlagregional, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = AufAbschlagRegional()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    def test_aufabschlagregional_betraege_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = (
                AufAbschlagRegional(
                    bezeichnung="foo",
                    betraege=[],
                ),
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



from datetime import datetime, timezone

import pytest
from pydantic import ValidationError

from bo4e.bo.ausschreibung import Ausschreibung
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.enum.ausschreibungsportal import Ausschreibungsportal
from bo4e.enum.ausschreibungsstatus import Ausschreibungsstatus
from bo4e.enum.ausschreibungstyp import Ausschreibungstyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_adresse import example_adresse
from tests.test_ausschreibungslos import example_ausschreibungslos
from tests.test_zeitraum import example_zeitraum


class TestAusschreibung:
    @pytest.mark.parametrize(
        "ausschreibung",
        [
            pytest.param(
                Ausschreibung(
                    ausschreibungsnummer="239230",
                    ausschreibungstyp=Ausschreibungstyp.PRIVATRECHTLICH,
                    ausschreibungsstatus=Ausschreibungsstatus.PHASE3,
                    kostenpflichtig=True,
                    ausschreibungportal=Ausschreibungsportal.BMWI,
                    webseite="https://meineausschreibungswebsite.inv/",
                    veroeffentlichungszeitpunkt=datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                    abgabefrist=example_zeitraum,
                    bindefrist=example_zeitraum,
                    ausschreibender=Geschaeftspartner(
                        name1="Batman",
                        gewerbekennzeichnung=True,
                        geschaeftspartnerrolle=[Geschaeftspartnerrolle.LIEFERANT],
                        partneradresse=example_adresse,
                    ),
                    lose=[example_ausschreibungslos],
                ),
                id="maximal attributes",
            ),
            pytest.param(
                Ausschreibung(
                    ausschreibungsnummer="239230",
                    ausschreibungstyp=Ausschreibungstyp.PRIVATRECHTLICH,
                    ausschreibungsstatus=Ausschreibungsstatus.PHASE3,
                    kostenpflichtig=True,
                    veroeffentlichungszeitpunkt=datetime(2022, 1, 1, 0, 0, 0, tzinfo=timezone.utc),
                    abgabefrist=example_zeitraum,
                    bindefrist=example_zeitraum,
                    ausschreibender=Geschaeftspartner(
                        name1="Batman",
                        gewerbekennzeichnung=True,
                        geschaeftspartnerrolle=[Geschaeftspartnerrolle.LIEFERANT],
                        partneradresse=example_adresse,
                    ),
                    lose=[example_ausschreibungslos],
                ),
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, ausschreibung: Ausschreibung) -> None:
        assert_serialization_roundtrip(ausschreibung)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Ausschreibung()  # type: ignore[call-arg]

        assert "9 validation errors" in str(excinfo.value)



import json

from bo4e.bo.marktteilnehmer import Marktteilnehmer
from bo4e.com.adresse import Adresse
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.marktrolle import Marktrolle
from bo4e.enum.rollencodetyp import Rollencodetyp
from bo4e.enum.sparte import Sparte

example_marktteilnehmer = Marktteilnehmer(
    # required attributes of Marktteilnehmer only
    marktrolle=Marktrolle.DL,
    rollencodenummer="9903916000000",
    rollencodetyp=Rollencodetyp.BDEW,
    sparte=Sparte.STROM,
    # required attributes inherited from Geschaeftspartner
    name1="Netze BW GmbH",
    gewerbekennzeichnung=True,
    geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
    partneradresse=Adresse(
        strasse="Schelmenwasenstraße",
        hausnummer="15",
        postleitzahl="70567",
        ort="Stuttgart",
    ),
)


class TestMarktteilnehmer:
    def test_serialization(self) -> None:
        mt = example_marktteilnehmer

        assert mt.versionstruktur == "2", "versionstruktur was not automatically set"
        assert mt.bo_typ == BoTyp.MARKTTEILNEHMER, "boTyp was not automatically set"

        json_string = mt.json(by_alias=True, ensure_ascii=False)
        json_dict = json.loads(json_string)

        # Test camelcase
        assert "boTyp" in json_dict
        assert "marktrolle" in json_dict

        deserialized_mt: Marktteilnehmer = Marktteilnehmer.parse_raw(json_string)

        assert mt.marktrolle is deserialized_mt.marktrolle
        # Test snakecase
        assert deserialized_mt.bo_typ is BoTyp.MARKTTEILNEHMER



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Wertermittlungsverfahren(StrEnum):
    """
    Gibt an, ob es sich um eine Prognose oder eine Messung handelt, beispielsweise bei der Abbildung eines Verbrauchs.
    """

    PROGNOSE = "PROGNOSE"  #: Prognose
    MESSUNG = "MESSUNG"  #: Messung



"""
Contains StandorteigenschaftenGas class
and corresponding marshmallow schema for de-/serialization
"""
from typing import List

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.com.com import COM
from bo4e.com.marktgebietinfo import MarktgebietInfo


class StandorteigenschaftenGas(COM):
    """
    Standorteigenschaften der Sparte Gas

    .. raw:: html

        <object data="../_static/images/bo4e/com/StandorteigenschaftenGas.svg" type="image/svg+xml"></object>

    .. HINT::
        `StandorteigenschaftenGas JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/StandorteigenschaftenGas.json>`_

    """

    # required attributes
    netzkontonummern: conlist(str, min_items=1, max_items=2)  # type: ignore[valid-type]  #: Netzkontonummern der Gasnetze
    marktgebiete: List[MarktgebietInfo]  #: Die Informationen zu Marktgebieten in dem Netz.



"""
Contains Tagesvektor class and corresponding marshmallow schema for de-/serialization
"""
import datetime

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.com.com import COM
from bo4e.com.zeitreihenwertkompakt import Zeitreihenwertkompakt


class Tagesvektor(COM):
    """
    Abbildung eines Tagesvektors eines beliebigen äquidistanten Zeitrasters

    .. raw:: html

        <object data="../_static/images/bo4e/com/Tagesvektor.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tagesvektor JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Tagesvektor.json>`_

    """

    # required attributes
    # for the validator see https://github.com/Hochfrequenz/BO4E-python/issues/261
    tag: datetime.datetime
    """
    Der Zeitpunkt ab dem die Werte ermittelt wurden.
    Es kann entweder der Beginn des Strom- oder Gastages verwendet werden.
    Der Zeitpunkt sollte eindeutig sein, d.h. sowohl Datum+Uhrzeit als auch den UTC-Offset spezifizieren.
    """
    # for the validator see also https://github.com/Hochfrequenz/BO4E-python/issues/262
    werte: conlist(Zeitreihenwertkompakt, min_items=1)  # type: ignore[valid-type]
    """
    Die Werte am angegebenen Tag;
    In Kombination aus Zeitintervall und Tag lassen sich die Zeiten der Werte eindeutig konstruieren.
    """



from datetime import datetime, timezone

import pytest
from pydantic import ValidationError

from bo4e.bo.angebot import Angebot
from bo4e.bo.ansprechpartner import Ansprechpartner
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.sparte import Sparte
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_adresse import example_adresse
from tests.test_angebotsvariante import example_angebotsvariante


class TestAngebot:
    @pytest.mark.parametrize(
        "angebot",
        [
            pytest.param(
                Angebot(
                    angebotsnummer="271828",
                    anfragereferenz="foo",
                    angebotsdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
                    sparte=Sparte.GAS,
                    bindefrist=datetime(2019, 3, 2, tzinfo=timezone.utc),
                    angebotsgeber=Geschaeftspartner(
                        name1="Batman",
                        gewerbekennzeichnung=True,
                        geschaeftspartnerrolle=[Geschaeftspartnerrolle.LIEFERANT],
                        partneradresse=example_adresse,
                    ),
                    angebotsnehmer=Geschaeftspartner(
                        name1="Joker",
                        gewerbekennzeichnung=False,
                        geschaeftspartnerrolle=[Geschaeftspartnerrolle.KUNDE],
                        partneradresse=example_adresse,
                    ),
                    unterzeichner_angebotsnehmer=Ansprechpartner(
                        nachname="Titans",
                        geschaeftspartner=Geschaeftspartner(
                            name1="Wonderwoman",
                            gewerbekennzeichnung=False,
                            geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
                            partneradresse=example_adresse,
                        ),
                    ),
                    unterzeichner_angebotsgeber=Ansprechpartner(
                        nachname="Titans",
                        geschaeftspartner=Geschaeftspartner(
                            name1="Robin",
                            gewerbekennzeichnung=False,
                            geschaeftspartnerrolle=[Geschaeftspartnerrolle.KUNDE],
                            partneradresse=example_adresse,
                        ),
                    ),
                    varianten=[example_angebotsvariante],
                ),
                id="required and optional attributes",
            ),
            pytest.param(
                Angebot(
                    angebotsnummer="271828",
                    angebotsdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
                    sparte=Sparte.GAS,
                    angebotsgeber=Geschaeftspartner(
                        name1="Batman",
                        gewerbekennzeichnung=True,
                        geschaeftspartnerrolle=[Geschaeftspartnerrolle.LIEFERANT],
                        partneradresse=example_adresse,
                    ),
                    angebotsnehmer=Geschaeftspartner(
                        name1="Joker",
                        gewerbekennzeichnung=False,
                        geschaeftspartnerrolle=[Geschaeftspartnerrolle.KUNDE],
                        partneradresse=example_adresse,
                    ),
                    varianten=[example_angebotsvariante],
                ),
                id="required attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, angebot: Angebot) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(angebot)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Angebot()  # type: ignore[call-arg]
        assert "6 validation errors" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class KundengruppeKA(StrEnum):
    """
    Eine Aufzählung zur Einordnung für die Höhe der Konzessionsabgabe.
    """

    S_SCHWACHLAST = "S_SCHWACHLAST"  #: Strom
    S_TARIF_25000 = "S_TARIF_25000"  #: Strom
    S_TARIF_100000 = "S_TARIF_100000"  #: Strom
    S_TARIF_500000 = "S_TARIF_500000"  #: Strom
    S_TARIF_G_500000 = "S_TARIF_G_500000"  #: Strom
    S_SONDERKUNDE = "S_SONDERKUNDE"  #: Strom
    G_KOWA_25000 = "G_KOWA_25000"  #: Gas
    G_KOWA_100000 = "G_KOWA_100000"  #: Gas
    G_KOWA_500000 = "G_KOWA_500000"  #: Gas
    G_KOWA_G_500000 = "G_KOWA_G_500000"  #: Gas
    G_TARIF_25000 = "G_TARIF_25000"  #: Gas
    G_TARIF_100000 = "G_TARIF_100000"  #: Gas
    G_TARIF_500000 = "G_TARIF_500000"  #: Gas
    G_TARIF_G_500000 = "G_TARIF_G_500000"  #: Gas
    G_SONDERKUNDE = "G_SONDERKUNDE"  #: Gas
    SONDER_KAS = "SONDER_KAS"  #: beides
    SONDER_SAS = "SONDER_SAS"  #: beides
    SONDER_TAS = "SONDER_TAS"  #: beides
    SONDER_TKS = "SONDER_TKS"  #: Gas
    SONDER_TSS = "SONDER_TSS"  #: Strom



"""
Contains Energiemix class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import List, Optional

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.com.com import COM
from bo4e.com.energieherkunft import Energieherkunft
from bo4e.enum.oekolabel import Oekolabel
from bo4e.enum.oekozertifikat import Oekozertifikat
from bo4e.enum.sparte import Sparte


class Energiemix(COM):
    """
    Zusammensetzung der gelieferten Energie aus den verschiedenen Primärenergieformen.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Energiemix.svg" type="image/svg+xml"></object>

    .. HINT::
        `Energiemix JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Energiemix.json>`_

    """

    # required attributes
    #: Eindeutige Nummer zur Identifizierung des Energiemixes
    energiemixnummer: int
    #: Strom oder Gas etc.
    energieart: Sparte
    #: Bezeichnung des Energiemix
    bezeichnung: str
    #: Jahr, für das der Energiemix gilt
    gueltigkeitsjahr: int
    #: Anteile der jeweiligen Erzeugungsart
    anteil: conlist(Energieherkunft, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Bemerkung zum Energiemix
    bemerkung: Optional[str] = None
    #: Höhe des erzeugten CO2-Ausstosses in g/kWh
    co2_emission: Optional[Decimal] = None
    #: Höhe des erzeugten Atommülls in g/kWh
    atommuell: Optional[Decimal] = None
    #: Zertifikate für den Energiemix
    oekozertifikate: List[Oekozertifikat] = []
    #: Ökolabel für den Energiemix
    oekolabel: List[Oekolabel] = []
    #: Kennzeichen, ob der Versorger zu den Öko Top Ten gehört
    oeko_top_ten: Optional[bool] = None
    #: Internetseite, auf der die Strommixdaten veröffentlicht sind
    website: Optional[str] = None



"""
Contains Zustaendigkeit class
and corresponding marshmallow schema for de-/serialization
"""
from typing import Optional

from bo4e.com.com import COM
from bo4e.enum.themengebiet import Themengebiet

# pylint: disable=too-few-public-methods


class Zustaendigkeit(COM):
    """
    Enthält die zeitliche Zuordnung eines Ansprechpartners zu Abteilungen und Zuständigkeiten.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zustaendigkeit.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zustaendigkeit JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Zustaendigkeit.json>`_

    """

    # required attributes
    themengebiet: Themengebiet
    """
    Hier kann eine thematische Zuordnung des Ansprechpartners angegeben werden
    """

    # optional attributes
    jobtitel: Optional[str] = None  #: Berufliche Rolle des Ansprechpartners
    abteilung: Optional[str] = None  #: Abteilung, in der der Ansprechpartner tätig ist



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


# pylint: disable=line-too-long
class Geraetetyp(StrEnum):
    """
    Auflistung möglicher abzurechnender Gerätetypen.
    """

    WECHSELSTROMZAEHLER = "WECHSELSTROMZAEHLER"  #: Wechselstromzähler
    DREHSTROMZAEHLER = "DREHSTROMZAEHLER"  #: Drehstromzähler
    ZWEIRICHTUNGSZAEHLER = "ZWEIRICHTUNGSZAEHLER"  #: Zweirichtungszähler
    RLM_ZAEHLER = "RLM_ZAEHLER"  #: RLM-Zähler
    BALGENGASZAEHLER = "BALGENGASZAEHLER"  #: Balgengaszähler
    MAXIMUMZAEHLER = "MAXIMUMZAEHLER"  #: Maximumzähler (Schleppzähler)
    MULTIPLEXANLAGE = "MULTIPLEXANLAGE"  #: Multiplexeranlage
    PAUSCHALANLAGE = "PAUSCHALANLAGE"  #: Pauschalanlagen
    VERSTAERKERANLAGE = "VERSTAERKERANLAGE"  #: Verstärkeranlage
    SUMMATIONSGERAET = "SUMMATIONSGERAET"  #: Summationsgerät
    IMPULSGEBER = "IMPULSGEBER"  #: Impulsgeber
    EDL_21_ZAEHLERAUFSATZ = "EDL_21_ZAEHLERAUFSATZ"  #: EDL 21 Zähleraufsatz für Zähler
    VIER_QUADRANTEN_LASTGANGZAEHLER = "VIER_QUADRANTEN_LASTGANGZAEHLER"  #: Vier-Quadranten-Lastgangzähler
    MENGENUMWERTER = "MENGENUMWERTER"  #: Mengenumwerter
    STROMWANDLER = "STROMWANDLER"  #: Stromwandler
    SPANNUNGSWANDLER = "SPANNUNGSWANDLER"  #: Spannungs-Wandler
    KOMBIMESSWANDLER = "KOMBIMESSWANDLER"  #: Kombimesswandler
    BLOCKSTROMWANDLER = "BLOCKSTROMWANDLER"  #: Blockstromwandler
    DATENLOGGER = "DATENLOGGER"  #: Datenlogger
    KOMMUNIKATIONSANSCHLUSS = "KOMMUNIKATIONSANSCHLUSS"  #: Kommunikationsanschluss
    MODEM = "MODEM"  #: Modem
    TELEKOMMUNIKATIONSEINRICHTUNG = "TELEKOMMUNIKATIONSEINRICHTUNG"  #: vom Messstellenbetreiber beigestellte Telekommunikationseinrichtung (Telefonanschluss)
    DREHKOLBENGASZAEHLER = "DREHKOLBENGASZAEHLER"  #: Drehkolbengaszähler
    TURBINENRADGASZAEHLER = "TURBINENRADGASZAEHLER"  #: Turbinenradgaszähler
    ULTRASCHALLZAEHLER = "ULTRASCHALLZAEHLER"  #: Ultraschallzähler
    WIRBELGASZAEHLER = "WIRBELGASZAEHLER"  #: Wirbelgaszähler
    MODERNE_MESSEINRICHTUNG = "MODERNE_MESSEINRICHTUNG"  #: moderne Messeinrichtung
    INTELLIGENTES_MESSYSTEM = "INTELLIGENTES_MESSYSTEM"  #: Intelligentes Messystem
    ELEKTRONISCHER_HAUSHALTSZAEHLER = "ELEKTRONISCHER_HAUSHALTSZAEHLER"  #: elektronischer Haushaltszähler
    STEUEREINRICHTUNG = "STEUEREINRICHTUNG"  #: Steuereinrichtung
    TARIFSCHALTGERAET = "TARIFSCHALTGERAET"  #: Tarifschaltgerät
    RUNDSTEUEREMPFAENGER = "RUNDSTEUEREMPFAENGER"  #: Rundsteuerempfänger
    OPTIONALE_ZUS_ZAEHLEINRICHTUNG = "OPTIONALE_ZUS_ZAEHLEINRICHTUNG"  #: optionale zusätzliche Zähleinrichtung
    MESSWANDLERSATZ_IMS_MME = "MESSWANDLERSATZ_IMS_MME"  #: Messwandlersatz Strom iMS und mME, NSP
    KOMBIMESSWANDLER_IMS_MME = "KOMBIMESSWANDLER_IMS_MME"  #: Kombimesswandlersatz (Strom u. Spg) iMS und mME
    TARIFSCHALTGERAET_IMS_MME = "TARIFSCHALTGERAET_IMS_MME"  #: Tarifschaltung iMS und mME
    RUNDSTEUEREMPFAENGER_IMS_MME = "RUNDSTEUEREMPFAENGER_IMS_MME"  #: Rundsteuerempfänger iMS und mME
    TEMPERATUR_KOMPENSATION = "TEMPERATUR_KOMPENSATION"  #: Temperaturkompensation
    HOECHSTBELASTUNGS_ANZEIGER = "HOECHSTBELASTUNGS_ANZEIGER"  #: Höchsbelastungsanzeiger
    SONSTIGES_GERAET = "SONSTIGES_GERAET"  #: Sonstiges Gerät
    PREPAYMENTZAEHLER = "PREPAYMENTZAEHLER"  #: Prepaymentzähler
    EDL_21 = "EDL_21"  #: EDL21
    _96_H_ZAEHLER = "96_H_ZAEHLER"  #: 96 h Zähler
    EDL_40_ZAEHLERAUFSATZ = "EDL_40_ZAEHLERAUFSATZ"  #: EDL 40 Zähleraufsatz für Zähler
    EDL_40 = "EDL_40"  #: EDL 40



from datetime import datetime

from bo4e.com.messlokationszuordnung import Messlokationszuordnung
from bo4e.enum.arithmetische_operation import ArithmetischeOperation


class TestMesslokationszuordnung:
    def test_serialisation_only_required_attributes(self) -> None:
        """
        Test serialisation of Messlokationszuordnung with required attributes only
        """
        messlokations_id = "DE0010688516810000000000000012345"

        mlz = Messlokationszuordnung(
            messlokations_id=messlokations_id,
            arithmetik=ArithmetischeOperation.ADDITION,
        )

        json_string = mlz.json(by_alias=True, ensure_ascii=False)

        assert messlokations_id in json_string
        assert "ADDITION" in json_string

        mlz_deserialized = Messlokationszuordnung.parse_raw(json_string)

        assert mlz_deserialized.messlokations_id == messlokations_id
        assert mlz_deserialized.arithmetik == ArithmetischeOperation.ADDITION

    def test_serialisation_required_and_optional_attributes(self) -> None:
        """
        Test serialisation of Messlokationszuordnung with required and optional attributes
        """
        messlokations_id = "DE0010688516810000000000000012345"

        mlz = Messlokationszuordnung(
            messlokations_id=messlokations_id,
            arithmetik=ArithmetischeOperation.ADDITION,
            gueltig_seit=datetime(year=2021, month=1, day=13),
            gueltig_bis=datetime(year=2021, month=5, day=4),
        )

        mlz_json = mlz.json(by_alias=True, ensure_ascii=False)

        # CamelCase keys are made because they will put into JSON strings
        # to send them to the frontend (= JavaScript land)
        assert "messlokationsId" in mlz_json
        assert messlokations_id in mlz_json
        assert "arithmetik" in mlz_json
        assert "ADDITION" in mlz_json
        assert "gueltigSeit" in mlz_json
        assert "2021-01-13T00:00:00" in mlz_json

        mlz_deserialized = Messlokationszuordnung.parse_raw(mlz_json)

        assert mlz_deserialized.messlokations_id == messlokations_id
        assert mlz_deserialized.arithmetik == ArithmetischeOperation.ADDITION
        assert mlz_deserialized.gueltig_seit == datetime(year=2021, month=1, day=13)
        assert mlz_deserialized.gueltig_bis == datetime(year=2021, month=5, day=4)



"""
Contains TarifpreisstaffelProOrt class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods


class TarifpreisstaffelProOrt(COM):
    """
    Gibt die Staffelgrenzen der jeweiligen Preise an

    .. raw:: html

        <object data="../_static/images/bo4e/com/TarifpreisstaffelProOrt.svg" type="image/svg+xml"></object>

    .. HINT::
        `TarifpreisstaffelProOrt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/TarifpreisstaffelProOrt.json>`_

    """

    # todo: decimal doesn't make sense here imo
    # https://github.com/Hochfrequenz/BO4E-python/issues/344

    # required attributes
    #: Der Arbeitspreis in ct/kWh
    arbeitspreis: Decimal
    #: Der Arbeitspreis für Verbräuche in der Niedertarifzeit in ct/kWh
    arbeitspreis_n_t: Decimal
    #: Der Grundpreis in Euro/Jahr
    grundpreis: Decimal
    #: Unterer Wert, ab dem die Staffel gilt (inklusive)
    staffelgrenze_von: Decimal
    #: Oberer Wert, bis zu dem die Staffel gilt (exklusive)
    staffelgrenze_bis: Decimal

    # there are no optional attributes



import sys

from setuptools import setup

min_version = (3, 8)

if sys.version_info < min_version:
    error = f"""
bo4e needs Python {min_version} or above.
You are using Python {sys.version_info}
""".format(
        ".".join(str(n) for n in min_version),
        ".".join(str(n) for n in sys.version_info[:3]),
    )
    sys.exit(error)

setup(name="bo4e", use_scm_version=True)
